# coding: utf-8

"""
LSF remote workflow implementation. See https://www.ibm.com/support/knowledgecenter/en/SSETD4_9.1.3.
"""

__all__ = ["LSFWorkflow"]


import contextlib
from abc import abstractmethod
from collections import OrderedDict

import luigi
import six

from law.config import Config
from law.workflow.remote import BaseRemoteWorkflow, BaseRemoteWorkflowProxy
from law.job.base import JobArguments, JobInputFile, DeprecatedInputFiles
from law.task.proxy import ProxyCommand
from law.target.file import get_path, get_scheme, FileSystemDirectoryTarget
from law.target.local import LocalDirectoryTarget
from law.parameter import NO_STR
from law.util import no_value, law_src_path, merge_dicts, DotDict
from law.logger import get_logger

from law.contrib.lsf.job import LSFJobManager, LSFJobFileFactory


logger = get_logger(__name__)


class LSFWorkflowProxy(BaseRemoteWorkflowProxy):

    workflow_type = "lsf"

    def create_job_manager(self, **kwargs):
        return self.task.lsf_create_job_manager(**kwargs)

    def create_job_file_factory(self, **kwargs):
        return self.task.lsf_create_job_file_factory(**kwargs)

    def create_job_file(self, job_num, branches):
        task = self.task

        # the file postfix is pythonic range made from branches, e.g. [0, 1, 2, 4] -> "_0To5"
        postfix = "_{}To{}".format(branches[0], branches[-1] + 1)

        # create the config
        c = self.job_file_factory.get_config()
        c.input_files = DeprecatedInputFiles()
        c.output_files = []
        c.render_variables = {}
        c.custom_content = []

        # get the actual wrapper file that will be executed by the remote job
        wrapper_file = task.lsf_wrapper_file()
        law_job_file = task.lsf_job_file()
        if wrapper_file and get_path(wrapper_file) != get_path(law_job_file):
            c.input_files["executable_file"] = wrapper_file
            c.executable = wrapper_file
        else:
            c.executable = law_job_file
        c.input_files["job_file"] = law_job_file

        # collect task parameters
        exclude_args = (
            task.exclude_params_branch |
            task.exclude_params_workflow |
            task.exclude_params_remote_workflow |
            task.exclude_params_lsf_workflow |
            {"workflow", "effective_workflow"}
        )
        proxy_cmd = ProxyCommand(
            task.as_branch(branches[0]),
            exclude_task_args=exclude_args,
            exclude_global_args=["workers", "local-scheduler", task.task_family + "-*"],
        )
        if task.lsf_use_local_scheduler():
            proxy_cmd.add_arg("--local-scheduler", "True", overwrite=True)
        for key, value in OrderedDict(task.lsf_cmdline_args()).items():
            proxy_cmd.add_arg(key, value, overwrite=True)

        # job script arguments
        job_args = JobArguments(
            task_cls=task.__class__,
            task_params=proxy_cmd.build(skip_run=True),
            branches=branches,
            workers=task.job_workers,
            auto_retry=False,
            dashboard_data=self.dashboard.remote_hook_data(
                job_num, self.job_data.attempts.get(job_num, 0)),
        )
        c.arguments = job_args.join()

        # add the bootstrap file
        bootstrap_file = task.lsf_bootstrap_file()
        if bootstrap_file:
            c.input_files["bootstrap_file"] = bootstrap_file

        # add the stageout file
        stageout_file = task.lsf_stageout_file()
        if stageout_file:
            c.input_files["stageout_file"] = stageout_file

        # does the dashboard have a hook file?
        dashboard_file = self.dashboard.remote_hook_file()
        if dashboard_file:
            c.input_files["dashboard_file"] = dashboard_file

        # initialize logs with empty values and defer to defaults later
        c.stdout = no_value
        c.stderr = no_value
        if task.transfer_logs:
            c.custom_log_file = "stdall.txt"

        # helper to cast directory paths to local directory targets if possible
        def cast_dir(output_dir, touch=True):
            if not isinstance(output_dir, FileSystemDirectoryTarget):
                path = get_path(output_dir)
                if get_scheme(path) not in (None, "file"):
                    return output_dir
                output_dir = LocalDirectoryTarget(path)
            if touch:
                output_dir.touch()
            return output_dir

        # when the output dir is local, we can run within this directory for easier output file
        # handling and use absolute paths for input files
        output_dir = cast_dir(task.lsf_output_directory())
        output_dir_is_local = isinstance(output_dir, LocalDirectoryTarget)
        if output_dir_is_local:
            c.absolute_paths = True
            c.cwd = output_dir.abspath

        # job name
        c.job_name = "{}{}".format(task.live_task_id, postfix)

        # task hook
        c = task.lsf_job_config(c, job_num, branches)

        # when the output dir is not local, direct output files are not possible
        if not output_dir_is_local:
            del c.output_files[:]

        # build the job file and get the sanitized config
        job_file, c = self.job_file_factory(postfix=postfix, **c.__dict__)

        # logging defaults
        # we do not use lsf's logging mechanism since it might require that the submission
        # directory is present when it retrieves logs, and therefore we use a custom log file
        c.stdout = c.stdout or None
        c.stderr = c.stderr or None
        c.custom_log_file = c.custom_log_file or None

        # get the location of the custom local log file if any
        abs_log_file = None
        if output_dir_is_local and c.custom_log_file:
            abs_log_file = output_dir.child(c.custom_log_file, type="f").abspath

        # return job and log files
        return {"job": job_file, "config": c, "log": abs_log_file}

    def destination_info(self):
        info = super(LSFWorkflowProxy, self).destination_info()

        if self.task.lsf_queue != NO_STR:
            info["queue"] = "queue: {}".format(self.task.lsf_queue)

        info = self.task.lsf_destination_info(info)

        return info


class LSFWorkflow(BaseRemoteWorkflow):

    workflow_proxy_cls = LSFWorkflowProxy

    lsf_workflow_run_decorators = None
    lsf_job_manager_defaults = None
    lsf_job_file_factory_defaults = None

    lsf_queue = luigi.Parameter(
        default=NO_STR,
        significant=False,
        description="target lsf queue; default: empty",
    )

    lsf_job_kwargs = ["lsf_queue"]
    lsf_job_kwargs_submit = None
    lsf_job_kwargs_cancel = None
    lsf_job_kwargs_query = None

    exclude_params_branch = {"lsf_queue"}

    exclude_params_lsf_workflow = set()

    exclude_index = True

    @contextlib.contextmanager
    def lsf_workflow_run_context(self):
        """
        Hook to provide a context manager in which the workflow run implementation is placed. This
        can be helpful in situations where resurces should be acquired before and released after
        running a workflow.
        """
        yield

    @abstractmethod
    def lsf_output_directory(self):
        """
        Hook to define the location of submission output files, such as the json files containing
        job data, and optional log files.
        This method should return a :py:class:`FileSystemDirectoryTarget`.
        """
        return None

    def lsf_bootstrap_file(self):
        return None

    def lsf_wrapper_file(self):
        return None

    def lsf_job_file(self):
        return JobInputFile(law_src_path("job", "law_job.sh"))

    def lsf_stageout_file(self):
        return None

    def lsf_workflow_requires(self):
        return DotDict()

    def lsf_output_postfix(self):
        return ""

    def lsf_job_resources(self, job_num, branches):
        """
        Hook to define resources for a specific job with number *job_num*, processing *branches*.
        This method should return a dictionary.
        """
        return {}

    def lsf_job_manager_cls(self):
        return LSFJobManager

    def lsf_create_job_manager(self, **kwargs):
        kwargs = merge_dicts(self.lsf_job_manager_defaults, kwargs)
        return self.lsf_job_manager_cls()(**kwargs)

    def lsf_job_file_factory_cls(self):
        return LSFJobFileFactory

    def lsf_create_job_file_factory(self, **kwargs):
        # get the file factory cls
        factory_cls = self.lsf_job_file_factory_cls()

        # job file fectory config priority: kwargs > class defaults
        kwargs = merge_dicts({}, self.lsf_job_file_factory_defaults, kwargs)

        # default mkdtemp value which might require task-level info
        if kwargs.get("mkdtemp") is None:
            cfg = Config.instance()
            mkdtemp = cfg.get_expanded(
                "job",
                cfg.find_option("job", "lsf_job_file_dir_mkdtemp", "job_file_dir_mkdtemp"),
            )
            if isinstance(mkdtemp, six.string_types) and mkdtemp.lower() not in {"true", "false"}:
                kwargs["mkdtemp"] = factory_cls._expand_template_path(
                    mkdtemp,
                    variables={"task_id": self.live_task_id, "task_family": self.task_family},
                )

        return factory_cls(**kwargs)

    def lsf_job_config(self, config, job_num, branches):
        return config

    def lsf_dump_intermediate_job_data(self):
        """
        Whether to dump intermediate job data to the job submission file while jobs are being
        submitted.
        """
        return True

    def lsf_post_submit_delay(self):
        """
        Configurable delay in seconds to wait after submitting jobs and before starting the status
        polling.
        """
        return self.poll_interval * 60

    def lsf_check_job_completeness(self):
        return False

    def lsf_check_job_completeness_delay(self):
        return 0.0

    def lsf_poll_callback(self, poll_data):
        """
        Configurable callback that is called after each job status query and before potential
        resubmission. It receives the variable polling attributes *poll_data* (:py:class:`PollData`)
        that can be changed within this method.

        If *False* is returned, the polling loop is gracefully terminated. Returning any other value
        does not have any effect.
        """
        return

    def lsf_use_local_scheduler(self):
        return True

    def lsf_cmdline_args(self):
        return {}

    def lsf_destination_info(self, info):
        return info



# coding: utf-8
# flake8: noqa

"""
Keras contrib functionality.
"""

__all__ = ["KerasModelFormatter", "KerasWeightsFormatter", "TFKerasModelFormatter"]


# provisioning imports
from law.contrib.keras.formatter import (
    KerasModelFormatter, KerasWeightsFormatter, TFKerasModelFormatter,
)



# coding: utf-8

"""
Collection of minor patches for luigi. These patches are only intended to support extra features for
law, rather than changing default luigi behavior.
"""

__all__ = ["before_run", "patch_all"]


import re
import functools
import copy
import logging

import luigi
import law
from law.logger import get_logger


logger = get_logger(__name__)

_patched = False

_before_run_funcs = []


def before_run(func, force=False):
    """
    Adds a function *func* to the list of callbacks that are invoked right before luigi starts
    running scheduled tasks. Unless *force* is *True*, a function that is already registered is not
    added again and *False* is returned. Otherwise, *True* is returned.
    """
    if func not in _before_run_funcs or force:
        _before_run_funcs.append(func)
        return True

    return False


def patch_all():
    """
    Runs all patches. This function ensures that a second invocation has no effect.
    """
    global _patched
    if _patched:
        return
    _patched = True

    patch_schedule_and_run()
    patch_task_process_run()
    patch_default_retcodes()
    patch_worker_add_task()
    patch_worker_add()
    patch_worker_run_task()
    patch_worker_get_work()
    patch_worker_factory()
    patch_keepalive_run()
    patch_luigi_run_result()
    patch_cmdline_parser()
    patch_interface_logging()
    patch_parameter_copy()
    patch_parameter_parse_or_no_value()
    patch_worker_check_complete_cached()

    logger.debug("applied all law-specific luigi patches")


def patch_schedule_and_run():
    """
    Patches ``luigi.interface._schedule_and_run`` to invoke all callbacks registered via
    :py:func:`before_run` right before luigi starts running scheduled tasks. This is achieved by
    patching ``luigi.worker.Worker.run`` within the scope of ``luigi.interface._schedule_and_run``.
    """
    _schedule_and_run_orig = luigi.interface._schedule_and_run

    @functools.wraps(_schedule_and_run_orig)
    def _schedule_and_run(*args, **kwargs):
        run_orig = luigi.worker.Worker.run

        @functools.wraps(run_orig)
        def run(self):
            # invoke all registered before_run functions
            for func in _before_run_funcs:
                if callable(func):
                    logger.debug("calling before_run function {}".format(func))
                    func()
                else:
                    logger.warning("registered before_run function {} is not callable".format(func))

            return run_orig(self)

        with law.util.patch_object(luigi.worker.Worker, "run", run):
            return _schedule_and_run_orig(*args, **kwargs)

    luigi.interface._schedule_and_run = _schedule_and_run

    logger.debug("patched luigi.interface._schedule_and_run")


def patch_task_process_run():
    """
    Patches ``luigi.worker.TaskProcess.run`` to increase the severity of luigi's interface logger
    when running local workflows that already yielded their branch tasks as dynamic dependencies.
    """
    run_orig = luigi.worker.TaskProcess.run

    interface_logger = logging.getLogger("luigi-interface")

    @functools.wraps(run_orig)
    def run(self):
        previous_level = interface_logger.level

        # update logging for local workflows that already yielded their branch tasks
        if (
            isinstance(self.task, law.LocalWorkflow) and
            self.task.is_workflow() and
            isinstance(self.task.workflow_proxy, law.workflow.local.LocalWorkflowProxy) and
            not self.task.local_workflow_require_branches and
            self.task.workflow_proxy._local_workflow_has_yielded
        ):
            interface_logger.setLevel(logging.WARNING)

        try:
            return run_orig(self)
        finally:
            interface_logger.setLevel(previous_level)

    luigi.worker.TaskProcess.run = run

    logger.debug("patched luigi.worker.TaskProcess.run")


def patch_default_retcodes():
    """
    Sets the default luigi return codes in ``luigi.retcodes.retcode`` to:

        - already_running: 10
        - missing_data: 20
        - not_run: 30
        - task_failed: 40
        - scheduling_error: 50
        - unhandled_exception: 60
    """
    import luigi.retcodes

    retcode = luigi.retcodes.retcode

    retcode.already_running._default = 10
    retcode.missing_data._default = 20
    retcode.not_run._default = 30
    retcode.task_failed._default = 40
    retcode.scheduling_error._default = 50
    retcode.unhandled_exception._default = 60

    logger.debug("patched luigis default return codes")


def patch_worker_add_task():
    """
    Patches the ``luigi.worker.Worker._add_task`` method to skip dependencies of the triggered task
    when running in a sandbox, as dependencies are already controlled from outside the sandbox.
    Also, the severity of luigi's interface logging is increased when running in a sandbox.
    """
    _add_task_orig = luigi.worker.Worker._add_task

    interface_logger = logging.getLogger("luigi-interface")

    @functools.wraps(_add_task_orig)
    def _add_task(self, *args, **kwargs):
        previous_level = interface_logger.level

        if law.sandbox.base._sandbox_switched:
            # increase the log level
            interface_logger.setLevel(logging.WARNING)

            # reset deps
            if "deps" in kwargs:
                kwargs["deps"] = None

        try:
            return _add_task_orig(self, *args, **kwargs)
        finally:
            interface_logger.setLevel(previous_level)

    luigi.worker.Worker._add_task = _add_task

    logger.debug("patched luigi.worker.Worker._add_task")


def patch_worker_add():
    """
    Patches the ``luigi.worker.Worker._add`` method to make sure that no dependencies are yielded
    when the triggered task is added to the worker when running in a sandbox and that the task is
    added to the scheduler with the id of the outer task.
    """
    _add_orig = luigi.worker.Worker._add

    @functools.wraps(_add_orig)
    def _add(self, task, *args, **kwargs):
        # _add_orig returns a generator, which we simply drain here
        # when we are in a sandbox
        if law.sandbox.base._sandbox_switched:
            task.task_id = law.sandbox.base._sandbox_task_id
            for _ in _add_orig(self, task, *args, **kwargs):
                pass
            return []
        else:
            return _add_orig(self, task, *args, **kwargs)

    luigi.worker.Worker._add = _add

    logger.debug("patched luigi.worker.Worker._add")


def patch_worker_run_task():
    """
    Patches the ``luigi.worker.Worker._run_task`` method to store the worker id and the id of its
    first task in the task. This information is required by the sandboxing mechanism.
    """
    _run_task_orig = luigi.worker.Worker._run_task

    @functools.wraps(_run_task_orig)
    def _run_task(self, task_id):
        task = self._scheduled_tasks[task_id]

        task._worker_id = self._id
        task._worker_first_task_id = self._first_task

        try:
            _run_task_orig(self, task_id)
        finally:
            task._worker_id = None
            task._worker_first_task_id = None

        # make worker disposable when sandboxed
        if law.sandbox.base._sandbox_switched:
            self._start_phasing_out()

    luigi.worker.Worker._run_task = _run_task

    logger.debug("patched luigi.worker.Worker._run_task")


def patch_worker_get_work():
    """
    Patches the ``luigi.worker.Worker._get_work`` method to only return information of the sandboxed
    task when running in a sandbox. This way, actual (outer) task and the sandboxed (inner) task
    appear to a central as the same task and communication for exchanging (e.g.) messages becomes
    transparent.
    """
    _get_work_orig = luigi.worker.Worker._get_work

    @functools.wraps(_get_work_orig)
    def _get_work(self):
        if law.sandbox.base._sandbox_switched:
            # when the worker is configured to stop requesting work, as triggered by the patched
            # _run_task method (see above), the worker response should contain an empty task_id
            task_id = None if self._stop_requesting_work else law.sandbox.base._sandbox_task_id
            return luigi.worker.GetWorkResponse(
                task_id=task_id,
                running_tasks=[],
                n_pending_tasks=0,
                n_unique_pending=0,
                n_pending_last_scheduled=0,
                worker_state=luigi.worker.WORKER_STATE_ACTIVE,
            )
        else:
            return _get_work_orig(self)

    luigi.worker.Worker._get_work = _get_work

    logger.debug("patched luigi.worker.Worker._get_work")


def patch_worker_factory():
    """
    Patches the ``luigi.interface._WorkerSchedulerFactory`` to include sandboxing information when
    creating a worker instance.
    """
    def create_worker(self, scheduler, worker_processes, assistant=False):
        worker = luigi.worker.Worker(scheduler=scheduler, worker_processes=worker_processes,
            assistant=assistant, worker_id=law.sandbox.base._sandbox_worker_id or None)
        worker._first_task = law.sandbox.base._sandbox_worker_first_task_id or None
        return worker

    luigi.interface._WorkerSchedulerFactory.create_worker = create_worker

    logger.debug("patched luigi.interface._WorkerSchedulerFactory.create_worker")


def patch_keepalive_run():
    """
    Patches the ``luigi.worker.KeepAliveThread.run`` to immediately stop the keep-alive thread when
    running within a sandbox.
    """
    run_orig = luigi.worker.KeepAliveThread.run

    @functools.wraps(run_orig)
    def run(self):
        # do not run the keep-alive loop when sandboxed
        if law.sandbox.base._sandbox_switched:
            self.stop()
        else:
            run_orig(self)

    luigi.worker.KeepAliveThread.run = run

    logger.debug("patched luigi.worker.KeepAliveThread.run")


def patch_luigi_run_result():
    __init__orig = luigi.execution_summary.LuigiRunResult.__init__

    @functools.wraps(__init__orig)
    def __init__(self, *args, **kwargs):
        __init__orig(self, *args, **kwargs)

        # condense the summary text into a single line when sandboxed
        if law.sandbox.base._sandbox_switched:
            self.summary_text = luigi.execution_summary._create_one_line_summary(self.status)

    luigi.execution_summary.LuigiRunResult.__init__ = __init__

    logger.debug("patched luigi.execution_summary.LuigiRunResult.__init__")


def patch_cmdline_parser():
    """
    Patches the ``luigi.cmdline_parser.CmdlineParser`` to store the original command line arguments
    for later processing in the :py:class:`law.config.Config`, and to update the way that parameter
    objects are called to parse empty strings.
    """
    __init__orig = luigi.cmdline_parser.CmdlineParser.__init__

    @functools.wraps(__init__orig)
    def __init__(self, cmdline_args):
        __init__orig(self, cmdline_args)
        self.cmdline_args = cmdline_args

    luigi.cmdline_parser.CmdlineParser.__init__ = __init__
    logger.debug("patched luigi.cmdline_parser.CmdlineParser.__init__")

    _get_task_kwargs_orig = luigi.cmdline_parser.CmdlineParser._get_task_kwargs

    @functools.wraps(_get_task_kwargs_orig)
    def _get_task_kwargs(self):
        res = {}
        for (param_name, param_obj) in self._get_task_cls().get_params():
            attr = getattr(self.known_args, param_name)
            # always skip None
            if attr is None:
                continue
            # skip other empty values unless the parameter has parse_empty set
            if not attr and not getattr(param_obj, "parse_empty", False):
                continue
            # parse the value
            res.update(((param_name, param_obj.parse(attr)),))

        return res

    luigi.cmdline_parser.CmdlineParser._get_task_kwargs = _get_task_kwargs
    logger.debug("patched luigi.cmdline_parser.CmdlineParser._get_task_kwargs")


def patch_interface_logging():
    """
    Patches ``luigi.setup_logging.InterfaceLogging._default`` to avoid adding multiple tty stream
    handlers to the logger named "luigi-interface" and to preserve any previously set log level.
    Also, the formatters of its stream handlers are amended in order to colorize parts of luigi log
    messages.
    """
    _default_orig = luigi.setup_logging.InterfaceLogging._default

    # predefined styles for luigi log messages
    sched_action_colors = {
        "PENDING": {"color": "cyan"},
        "DONE": {"color": "green"},
        "FAILED": {"color": "red"},
    }
    sched_task_style = {"style": "bright"}
    sched_done_style = {"color": "green", "style": "bright"}
    worker_action_colors = {
        "running": {"color": "cyan"},
        "done": {"color": "green"},
        "failed": {"color": "red"},
        "new requirements": {"color": "magenta"},
    }
    worker_task_style = sched_task_style

    # precompiled expressions
    cre_sched_action = r"^(Informed scheduler that task\s+)([^\s]+)(\s+has\sstatus\s+)({})(.*)$"
    cre_sched_action = re.compile(cre_sched_action.format("|".join(sched_action_colors.keys())))
    cre_sched_done = re.compile("^(Done scheduling tasks)$")
    cre_worker_action = r"^(\[pid \d+\] Worker Worker\(.+\)\s+)({})(\s+)([^\(]+)(\(.+)$"
    cre_worker_action = re.compile(cre_worker_action.format("|".join(worker_action_colors.keys())))

    # log message formatter to partially colorize some luigi logs
    def colorize_luigi_logs(record):
        msg = record.getMessage()

        # worker task messages
        m = cre_worker_action.match(msg)
        if m:
            s1, action, s2, task, s3 = m.groups()
            task = law.util.colored(task, **worker_task_style)
            action = law.util.colored(action, **worker_action_colors[action])
            return s1 + action + s2 + task + s3

        # scheduler task registration messages
        m = cre_sched_action.match(msg)
        if m:
            s1, task, s2, action, s3 = m.groups()
            task = law.util.colored(task, **sched_task_style)
            action = law.util.colored(action, **sched_action_colors[action])
            return s1 + task + s2 + action + s3

        # scheduler done building tree
        m = cre_sched_done.match(msg)
        if m:
            msg = law.util.colored(m.group(1), **sched_done_style)
            return msg

        return msg

    @functools.wraps(_default_orig)
    def _default(cls, opts):
        _logger = logging.getLogger("luigi-interface")

        level_before = _logger.level
        tty_handlers_before = law.logger.get_tty_handlers(_logger)

        ret = _default_orig(opts)

        level_after = _logger.level
        tty_handlers_after = law.logger.get_tty_handlers(_logger)

        if level_before != logging.NOTSET and level_before != level_after:
            _logger.setLevel(level_before)

        if tty_handlers_before:
            for handler in tty_handlers_after[len(tty_handlers_before):]:
                _logger.removeHandler(handler)

        # update formatters to colorize messages
        for handler in _logger.handlers:
            if isinstance(handler.formatter, law.logger.LogFormatter):
                handler.formatter.format_msg = colorize_luigi_logs

        return ret

    luigi.setup_logging.InterfaceLogging._default = classmethod(_default)

    logger.debug("patched luigi.setup_logging.InterfaceLogging._default")


def patch_parameter_copy():
    """
    Patches ``luigi.parameter.Parameter`` to add a convenience methods that allows to copy parameter
    instances and assigning new attributes such as descriptions or default values. This same
    functionality will eventually be moved to luigi, but the patch might be kept for versions of
    luigi where it was not addded yet.
    """
    default_cre = re.compile(r"(.+)(;|,)\s*((empty|no|without) default|default: [^\;]+)\s*$")

    def _copy(self, add_default_to_description=False, **kwargs):
        # copy the instance
        inst = copy.copy(self)

        # kwargs should in general match those accepted in the constructor, which are mostly saved
        # as instance attributes using the same name, except in some cases which must be redirected
        if "default" in kwargs and "_default" not in kwargs:
            kwargs["_default"] = kwargs.pop("default")
        if "config_path" in kwargs and "_config_path" not in kwargs:
            kwargs["_config_path"] = kwargs.pop("config_path")

        # overwrite attributes
        inst.__dict__.update(kwargs)

        # amend the description
        if add_default_to_description:
            # remove default from description
            if inst.description:
                m = default_cre.match(inst.description)
                if m:
                    inst.description = m.group(1)
                inst.description += "; "
            inst.description += "default: {}".format(inst._default)

        return inst

    luigi.parameter.Parameter.copy = _copy

    logger.debug("patched luigi.parameter.Parameter.copy")


def patch_parameter_parse_or_no_value():
    """
    Patches ``luigi.parameter.Parameter`` to properly accept empty values such as empty strings for
    normal parameters or zeros for integer parameters instead of treating them as missing and to be
    replaced with default values.
    """
    def _parse_or_no_value(self, x):
        empty = x is None or x is luigi.parameter._no_value
        return luigi.parameter._no_value if empty else self.parse(x)

    luigi.parameter.Parameter._parse_or_no_value = _parse_or_no_value

    logger.debug("patched luigi.parameter.Parameter._parse_or_no_value")


def patch_worker_check_complete_cached():
    """
    Patches the ``luigi.worker.check_complete_cached`` function to treat cached task completeness
    decision slightly differently. The original implementation only skips the completeness check and
    uses the cached value if, and only if, a task was actually already marked as complete. Missing
    or *False* entries are both neglected and the completeness check is performed. Now, *False*
    entries also cause the check to be skipped, considering the task as incomplete. However, after
    that, the cache entry is removed so that subsequent checks are performed as usual.
    """
    def check_complete_cached(task, completion_cache=None):
        # no caching behavior when no cache is given
        if completion_cache is None:
            return task.complete()

        # get the cached state
        cache_key = task.task_id
        complete = completion_cache.get(cache_key)

        # stop when already complete
        if complete:
            return True

        # consider as incomplete when the cache entry is falsy, yet not None
        if not complete and complete is not None:
            completion_cache.pop(cache_key, None)
            return False

        # check the status and tell the cache when complete
        complete = task.complete()
        if complete:
            completion_cache[cache_key] = complete

        return complete

    luigi.worker.check_complete_cached = check_complete_cached

    logger.debug("patched luigi.worker.check_complete_cached")



# coding: utf-8

"""
Law command line interface entry point.
"""


import sys
from importlib import import_module
from argparse import ArgumentParser

import law


progs = ["run", "index", "config", "software", "completion", "location", "quickstart", "luigid"]


def run(argv=None):
    """
    Entry point to the law cli. Sets up all parsers, parses all arguments given by *argv*, and
    executes the requested subprogram. When *None*, *argv* defaults to ``sys.argv[1:]``.
    """
    # setup the main parser and sub parsers
    parser = ArgumentParser(
        prog="law",
        description="The law command line tool.",
    )
    sub_parsers = parser.add_subparsers(
        help="subcommands",
        dest="command",
    )

    # add main arguments
    parser.add_argument(
        "--version",
        "-V",
        action="version",
        version=law.__version__,
    )

    # setup all progs
    mods = {}
    for prog in progs:
        mods[prog] = import_module("law.cli." + prog)
        mods[prog].setup_parser(sub_parsers)

    # default argv
    if not argv:
        argv = sys.argv[1:]

    # argv that is passed to the prog execution when set
    prog_argv = None

    # parse args and dispatch execution, considering some special cases
    prog = argv[0] if argv else None
    if prog == "run":
        args = parser.parse_args(argv[:2])
        prog_argv = ["law run"] + argv
    elif prog == "luigid":
        args = parser.parse_args(argv[:1])
        prog_argv = ["law luigid"] + argv
    else:
        args = parser.parse_args(argv)

    # the parser determines the prog, so overwrite it
    prog = args.command
    if not prog:
        parser.print_help()
        return 0

    exec_args = (args,)
    if prog_argv is not None:
        exec_args += (prog_argv,)

    return mods[prog].execute(*exec_args)



# coding: utf-8

"""
"law quickstart" cli subprogram.
"""


import os
import shutil

from law.config import Config
from law.util import law_src_path


_cfg = Config.instance()


def setup_parser(sub_parsers):
    """
    Sets up the command line parser for the *quickstart* subprogram and adds it to *sub_parsers*.
    """
    parser = sub_parsers.add_parser(
        "quickstart",
        prog="law quickstart",
        description="Quickstart to create a minimal project structure and law configuration.",
    )

    parser.add_argument(
        "--directory",
        "-d",
        help="the directory where the quickstart files are created; default: current directory",
    )
    parser.add_argument(
        "--no-tasks",
        action="store_true",
        help="skip creating tasks",
    )
    parser.add_argument(
        "--no-config",
        action="store_true",
        help="skip creating the law.cfg file",
    )
    parser.add_argument(
        "--no-setup",
        action="store_true",
        help="skip creating the setup.sh file",
    )


def execute(args):
    """
    Executes the *quickstart* subprogram with parsed commandline *args*.
    """
    # get the quickstart directory
    qs_dir = law_src_path("templates", "quickstart")

    # prepare the directory if it does not exist yet
    out_dir = os.path.normpath(os.path.abspath(args.directory))
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)

    # copy tasks
    if not args.no_tasks:
        dst = os.path.join(out_dir, "my_package")
        shutil.copytree(os.path.join(qs_dir, "my_package"), dst)
        print("created {}".format(dst))

    # copy config
    if not args.no_config:
        shutil.copy2(os.path.join(qs_dir, "law.cfg"), out_dir)
        print("created {}".format(os.path.join(out_dir, "law.cfg")))

    # copy setup
    if not args.no_setup:
        shutil.copy2(os.path.join(qs_dir, "setup.sh"), out_dir)
        print("created {}".format(os.path.join(out_dir, "setup.sh")))



# coding: utf-8

"""
HTCondor workflow implementation. See https://research.cs.wisc.edu/htcondor.
"""

__all__ = ["HTCondorWorkflow"]


import os
import contextlib
from abc import abstractmethod
from collections import OrderedDict

import luigi
import six

from law.config import Config
from law.workflow.remote import BaseRemoteWorkflow, BaseRemoteWorkflowProxy
from law.job.base import JobArguments, JobInputFile
from law.task.proxy import ProxyCommand
from law.target.file import get_path, get_scheme, FileSystemDirectoryTarget
from law.target.local import LocalDirectoryTarget
from law.parameter import NO_STR
from law.util import no_value, law_src_path, merge_dicts, DotDict, rel_path
from law.logger import get_logger

from law.contrib.htcondor.job import HTCondorJobManager, HTCondorJobFileFactory


logger = get_logger(__name__)


class HTCondorWorkflowProxy(BaseRemoteWorkflowProxy):

    workflow_type = "htcondor"

    def create_job_manager(self, **kwargs):
        return self.task.htcondor_create_job_manager(**kwargs)

    def create_job_file_factory(self, **kwargs):
        return self.task.htcondor_create_job_file_factory(**kwargs)

    def create_job_file(self, *args):
        task = self.task

        grouped_submission = len(args) == 1
        if grouped_submission:
            submit_jobs = args[0]
        else:
            job_num, branches = args

        # create the config
        c = self.job_file_factory.get_config()
        c.input_files = {}
        c.output_files = {}
        c.render_variables = {}
        c.custom_content = []

        # get the actual wrapper and job file that will be executed by the remote job
        law_job_file = task.htcondor_job_file()
        c.input_files["job_file"] = law_job_file
        if grouped_submission:
            # grouped wrapper file
            wrapper_file = task.htcondor_group_wrapper_file()
            c.input_files["executable_file"] = wrapper_file
            c.executable = wrapper_file
        else:
            # standard wrapper file
            wrapper_file = task.htcondor_wrapper_file()
            if wrapper_file and get_path(wrapper_file) != get_path(law_job_file):
                c.input_files["executable_file"] = wrapper_file
                c.executable = wrapper_file
            else:
                c.executable = law_job_file

        # collect task parameters
        exclude_args = (
            task.exclude_params_branch |
            task.exclude_params_workflow |
            task.exclude_params_remote_workflow |
            task.exclude_params_htcondor_workflow |
            {"workflow", "effective_workflow"}
        )
        proxy_cmd = ProxyCommand(
            task.as_branch(0 if grouped_submission else branches[0]),
            exclude_task_args=exclude_args,
            exclude_global_args=["workers", "local-scheduler", task.task_family + "-*"],
        )
        if task.htcondor_use_local_scheduler():
            proxy_cmd.add_arg("--local-scheduler", "True", overwrite=True)
        for key, value in OrderedDict(task.htcondor_cmdline_args()).items():
            proxy_cmd.add_arg(key, value, overwrite=True)

        # the file postfix is pythonic range made from branches, e.g. [0, 1, 2, 4] -> "_0To5"
        if grouped_submission:
            c.postfix = [
                "_{}To{}".format(branches[0], branches[-1] + 1)
                for branches in submit_jobs.values()
            ]
        else:
            c.postfix = "_{}To{}".format(branches[0], branches[-1] + 1)

        # job script arguments per job number
        def get_job_args(job_num, branches):
            return JobArguments(
                task_cls=task.__class__,
                task_params=proxy_cmd.build(skip_run=True),
                branches=branches,
                workers=task.job_workers,
                auto_retry=False,
                dashboard_data=self.dashboard.remote_hook_data(
                    job_num, self.job_data.attempts.get(job_num, 0)),
            )

        if grouped_submission:
            c.arguments = [
                get_job_args(job_num, branches).join()
                for job_num, branches in submit_jobs.items()
            ]
        else:
            c.arguments = get_job_args(job_num, branches).join()

        # add the bootstrap file
        bootstrap_file = task.htcondor_bootstrap_file()
        if bootstrap_file:
            c.input_files["bootstrap_file"] = bootstrap_file

        # add the stageout file
        stageout_file = task.htcondor_stageout_file()
        if stageout_file:
            c.input_files["stageout_file"] = stageout_file

        # does the dashboard have a hook file?
        dashboard_file = self.dashboard.remote_hook_file()
        if dashboard_file:
            c.input_files["dashboard_file"] = dashboard_file

        # initialize logs with empty values and defer to defaults later
        c.log = no_value
        c.stdout = no_value
        c.stderr = no_value
        if task.transfer_logs:
            c.custom_log_file = "stdall.txt"

        # helper to cast directory paths to local directory targets if possible
        def cast_dir(output_dir, touch=True):
            if not isinstance(output_dir, FileSystemDirectoryTarget):
                path = get_path(output_dir)
                if get_scheme(path) not in (None, "file"):
                    return output_dir
                output_dir = LocalDirectoryTarget(path)
            if touch:
                output_dir.touch()
            return output_dir

        # when the output dir is local, we can run within this directory for easier output file
        # handling and use absolute paths for input files
        output_dir = cast_dir(task.htcondor_output_directory())
        output_dir_is_local = isinstance(output_dir, LocalDirectoryTarget)
        if output_dir_is_local:
            c.absolute_paths = True
            c.custom_content.append(("initialdir", output_dir.abspath))

        # prepare the log dir
        log_dir_orig = task.htcondor_log_directory()
        log_dir = cast_dir(log_dir_orig) if log_dir_orig else output_dir
        log_dir_is_local = isinstance(log_dir, LocalDirectoryTarget)

        # task hook
        if grouped_submission:
            c = task.htcondor_job_config(c, list(submit_jobs.keys()), list(submit_jobs.values()))
        else:
            c = task.htcondor_job_config(c, job_num, branches)

        # logging defaults
        # we do not use htcondor's logging mechanism since it might require that the submission
        # directory is present when it retrieves logs, and therefore we use a custom log file
        # also, stderr and stdout can be remapped (moved) by htcondor, so use a different behavior
        def log_path(path):
            if not path:
                return None
            log_target = log_dir.child(path, type="f")
            if log_target.parent != log_dir:
                log_target.parent.touch()
            return log_target.abspath

        c.log = c.log or None
        c.stdout = log_path(c.stdout)
        c.stderr = log_path(c.stderr)
        c.custom_log_file = log_path(c.custom_log_file)

        # when the output dir is not local, direct output files are not possible
        if not output_dir_is_local and c.output_files:
            c.output_files.clear()

        # build the job file and get the sanitized config
        job_file, c = self.job_file_factory(grouped_submission=grouped_submission, **c.__dict__)

        # get the finale, absolute location of the custom log file
        # (note that c.custom_log_file is always just a basename after the factory hook)
        abs_log_file = None
        if log_dir_is_local and c.custom_log_file:
            abs_log_file = os.path.join(log_dir.abspath, c.custom_log_file)

        # return job and log files
        return {"job": job_file, "config": c, "log": abs_log_file}

    def _submit_group(self, *args, **kwargs):
        job_ids, submission_data = super(HTCondorWorkflowProxy, self)._submit_group(*args, **kwargs)

        # when a log file is present, replace certain htcondor variables
        for i, (job_id, (job_num, data)) in enumerate(zip(job_ids, submission_data.items())):
            # skip exceptions
            if isinstance(job_id, Exception):
                continue
            log = data.get("log")
            if not log:
                continue
            log_orig = log
            # replace Cluster, ClusterId, Process, ProcId
            c, p = job_id.split(".")
            log = log.replace("$(Cluster)", c).replace("$(ClusterId)", c)
            log = log.replace("$(Process)", p).replace("$(ProcId)", p)
            # replace law_job_postfix
            if data["config"].postfix_output_files and data["config"].postfix:
                log = log.replace("$(law_job_postfix)", data["config"].postfix[i])
            # nothing to do when the log did not changed
            if log == log_orig:
                continue
            # add back in a shallow copy
            data = data.copy()
            data["log"] = log
            submission_data[job_num] = data

        return job_ids, submission_data

    def destination_info(self):
        info = super(HTCondorWorkflowProxy, self).destination_info()

        if self.task.htcondor_pool and self.task.htcondor_pool != NO_STR:
            info["pool"] = "pool: {}".format(self.task.htcondor_pool)

        if self.task.htcondor_scheduler and self.task.htcondor_scheduler != NO_STR:
            info["scheduler"] = "scheduler: {}".format(self.task.htcondor_scheduler)

        info = self.task.htcondor_destination_info(info)

        return info


class HTCondorWorkflow(BaseRemoteWorkflow):

    workflow_proxy_cls = HTCondorWorkflowProxy

    htcondor_workflow_run_decorators = None
    htcondor_job_manager_defaults = None
    htcondor_job_file_factory_defaults = None

    htcondor_pool = luigi.Parameter(
        default=NO_STR,
        significant=False,
        description="target htcondor pool; default: empty",
    )
    htcondor_scheduler = luigi.Parameter(
        default=NO_STR,
        significant=False,
        description="target htcondor scheduler; default: empty",
    )

    htcondor_job_kwargs = ["htcondor_pool", "htcondor_scheduler"]
    htcondor_job_kwargs_submit = None
    htcondor_job_kwargs_cancel = None
    htcondor_job_kwargs_query = None

    exclude_params_branch = {"htcondor_pool", "htcondor_scheduler"}

    exclude_params_htcondor_workflow = set()

    exclude_index = True

    @contextlib.contextmanager
    def htcondor_workflow_run_context(self):
        """
        Hook to provide a context manager in which the workflow run implementation is placed. This
        can be helpful in situations where resurces should be acquired before and released after
        running a workflow.
        """
        yield

    @abstractmethod
    def htcondor_output_directory(self):
        """
        Hook to define the location of submission output files, such as the json files containing
        job data, and optional log files (in case :py:meth:`htcondor_log_directory` is not defined).
        This method should return a :py:class:`FileSystemDirectoryTarget`.
        """
        return None

    def htcondor_log_directory(self):
        """
        Hook to define the location of log files if any are written. When set, it has precedence
        over :py:meth:`htcondor_output_directory` for log files.
        This method should return a :py:class:`FileSystemDirectoryTarget` or a value that evaluates
        to *False* in case no custom log directory is desired.
        """
        return None

    def htcondor_workflow_requires(self):
        return DotDict()

    def htcondor_job_resources(self, job_num, branches):
        """
        Hook to define resources for a specific job with number *job_num*, processing *branches*.
        This method should return a dictionary.
        """
        return {}

    def htcondor_bootstrap_file(self):
        return None

    def htcondor_group_wrapper_file(self):
        # only used for grouped submissions
        return JobInputFile(
            path=rel_path(__file__, "htcondor_wrapper.sh"),
            copy=True,
            render_local=True,
            increment=True,
        )

    def htcondor_wrapper_file(self):
        return None

    def htcondor_job_file(self):
        return JobInputFile(
            path=law_src_path("job", "law_job.sh"),
            copy=True,
            share=True,
            render_job=True,
        )

    def htcondor_stageout_file(self):
        return None

    def htcondor_output_postfix(self):
        return ""

    def htcondor_job_manager_cls(self):
        return HTCondorJobManager

    def htcondor_create_job_manager(self, **kwargs):
        kwargs = merge_dicts(self.htcondor_job_manager_defaults, kwargs)
        return self.htcondor_job_manager_cls()(**kwargs)

    def htcondor_job_file_factory_cls(self):
        return HTCondorJobFileFactory

    def htcondor_create_job_file_factory(self, **kwargs):
        # get the file factory cls
        factory_cls = self.htcondor_job_file_factory_cls()

        # job file fectory config priority: kwargs > class defaults
        kwargs = merge_dicts({}, self.htcondor_job_file_factory_defaults, kwargs)

        # default mkdtemp value which might require task-level info
        if kwargs.get("mkdtemp") is None:
            cfg = Config.instance()
            mkdtemp = cfg.get_expanded(
                "job",
                cfg.find_option("job", "htcondor_job_file_dir_mkdtemp", "job_file_dir_mkdtemp"),
            )
            if isinstance(mkdtemp, six.string_types) and mkdtemp.lower() not in {"true", "false"}:
                kwargs["mkdtemp"] = factory_cls._expand_template_path(
                    mkdtemp,
                    variables={"task_id": self.live_task_id, "task_family": self.task_family},
                )

        return factory_cls(**kwargs)

    def htcondor_job_config(self, config, job_num, branches):
        return config

    def htcondor_dump_intermediate_job_data(self):
        """
        Whether to dump intermediate job data to the job submission file while jobs are being
        submitted.
        """
        return True

    def htcondor_post_submit_delay(self):
        """
        Configurable delay in seconds to wait after submitting jobs and before starting the status
        polling.
        """
        return self.poll_interval * 60

    def htcondor_check_job_completeness(self):
        return False

    def htcondor_check_job_completeness_delay(self):
        return 0.0

    def htcondor_poll_callback(self, poll_data):
        """
        Configurable callback that is called after each job status query and before potential
        resubmission. It receives the variable polling attributes *poll_data* (:py:class:`PollData`)
        that can be changed within this method.

        If *False* is returned, the polling loop is gracefully terminated. Returning any other value
        does not have any effect.
        """
        return

    def htcondor_use_local_scheduler(self):
        return False

    def htcondor_cmdline_args(self):
        return {}

    def htcondor_destination_info(self, info):
        return info



# coding: utf-8

"""
Function returning the config defaults of the gfal package.
"""


def config_defaults(default_config):
    return {
        "logging": {
            "gfal2": "WARNING",
        },
    }



# coding: utf-8

"""
Slack-related parameters.
"""

__all__ = ["NotifySlackParameter"]


import collections

import six

from law.parameter import NotifyParameter
from law.util import escape_markdown
from law.contrib.slack.notification import notify_slack


class NotifySlackParameter(NotifyParameter):

    def __init__(self, *args, **kwargs):
        super(NotifySlackParameter, self).__init__(*args, **kwargs)

        if not self.description:
            self.description = "when true, and the task's run method is decorated with " \
                "law.decorator.notify, a Slack notification is sent once the task finishes"

    @staticmethod
    def notify(success, title, content, **kwargs):
        # escape the full content
        content = content.__class__(
            (k, escape_markdown(v) if isinstance(v, six.string_types) else v)
            for k, v in content.items()
        )

        # overwrite title with slack markdown markup
        title = "*Notification from* `{}`".format(content["Task"])
        del content["Task"]

        # markup for traceback
        if "Traceback" in content:
            content["Traceback"] = "```{}```".format(content["Traceback"])

        # prepend the status text to the message content
        # emojis are "party popper" and "exclamation mark"
        parts = list(content.items())
        status_text = "success \xF0\x9F\x8E\x89" if success else "failure \xE2\x9D\x97"
        parts.insert(0, ("Status", status_text))
        content = collections.OrderedDict(parts)

        # attachment color depends on success
        color = "#4bb543" if success else "#ff0033"

        # send the notification
        return notify_slack(title, content, attachment_color=color, **kwargs)

    def get_transport(self):
        return {
            "func": self.notify,
            "raw": True,
            "colored": False,
        }



# coding: utf-8
# flake8: noqa

"""
Targets providing functionality to access files on Dropbox.
"""

__all__ = ["DropboxFileSystem", "DropboxTarget", "DropboxFileTarget", "DropboxDirectoryTarget"]


# provisioning imports
from law.contrib.dropbox.target import (
    DropboxFileSystem, DropboxTarget, DropboxFileTarget, DropboxDirectoryTarget,
)



# coding: utf-8

"""
CMS CRAB remote workflow implementation. See
https://twiki.cern.ch/twiki/bin/view/CMSPublic/SWGuideCrab.
"""

__all__ = ["CrabWorkflow"]


import uuid
import contextlib
from abc import abstractmethod
from collections import OrderedDict

import six

import law
from law.config import Config
from law.workflow.remote import BaseRemoteWorkflow, BaseRemoteWorkflowProxy
from law.job.base import JobArguments, JobInputFile
from law.target.file import get_path, get_scheme, remove_scheme, FileSystemDirectoryTarget
from law.target.local import LocalDirectoryTarget
from law.task.proxy import ProxyCommand
from law.util import no_value, law_src_path, merge_dicts, DotDict, human_duration
from law.logger import get_logger

from law.contrib.cms.job import CrabJobManager, CrabJobFileFactory
from law.contrib.cms.util import renew_vomsproxy, delegate_myproxy


logger = get_logger(__name__)


class CrabWorkflowProxy(BaseRemoteWorkflowProxy):

    workflow_type = "crab"

    # job script error codes are not transferred, so disable them
    job_error_messages = {}

    def create_job_manager(self, **kwargs):
        return self.task.crab_create_job_manager(**kwargs)

    def setup_job_manager(self):
        cfg = Config.instance()
        password_file = cfg.get_expanded("job", "crab_password_file")

        # ensure a VOMS proxy exists
        if not law.wlcg.check_vomsproxy_validity():
            print("renew voms-proxy")
            renew_vomsproxy(password_file=password_file)

        # ensure that it has been delegated to the myproxy server
        info = law.wlcg.get_myproxy_info(silent=True)
        delegate = False
        if not info:
            delegate = True
        elif "username" not in info:
            logger.warning("field 'username' not in myproxy info")
            delegate = True
        elif "timeleft" not in info:
            logger.warning("field 'timeleft' not in myproxy info")
            delegate = True
        elif info["timeleft"] < 86400:
            logger.warning("myproxy lifetime below 24h ({})".format(
                human_duration(seconds=info["timeleft"]),
            ))
            delegate = True

        # actual delegation
        if delegate:
            print("delegate to myproxy server")
            myproxy_username = delegate_myproxy(password_file=password_file)
        else:
            myproxy_username = info["username"]

        return {"myproxy_username": myproxy_username}

    def create_job_file_factory(self, **kwargs):
        return self.task.crab_create_job_file_factory(**kwargs)

    def create_job_file(self, submit_jobs):
        task = self.task

        # create the config
        c = self.job_file_factory.get_config()
        c.input_files = {}
        c.output_files = []
        c.render_variables = {}
        c.custom_content = []

        # get remote job file, force remote rendering
        law_job_file = JobInputFile(task.crab_job_file())
        law_job_file = JobInputFile(law_job_file.path, copy=False, render_job=True)
        c.executable = law_job_file
        c.input_files["job_file"] = c.executable

        # collect task parameters
        exclude_args = (
            task.exclude_params_branch |
            task.exclude_params_workflow |
            task.exclude_params_remote_workflow |
            task.exclude_params_crab_workflow |
            {"workflow", "effective_workflow"}
        )
        proxy_cmd = ProxyCommand(
            task.as_branch(),
            exclude_task_args=exclude_args,
            exclude_global_args=["workers", task.task_family + "-*"],
        )
        proxy_cmd.add_arg("--local-scheduler", "True", overwrite=True)
        for key, value in OrderedDict(task.crab_cmdline_args()).items():
            proxy_cmd.add_arg(key, value, overwrite=True)

        # job script arguments per job number
        c.arguments = []
        for job_num, branches in submit_jobs.items():
            job_args = JobArguments(
                task_cls=task.__class__,
                task_params=proxy_cmd.build(skip_run=True),
                branches=branches,
                workers=task.job_workers,
                auto_retry=False,
                dashboard_data=self.dashboard.remote_hook_data(
                    job_num, self.job_data.attempts.get(job_num, 0)),
            )
            c.arguments.append(job_args.join())

        # add the work area
        c.work_area = get_path(task.crab_work_area())

        # add the request name
        c.request_name = task.crab_request_name(submit_jobs).replace(".", "_")

        # add the storage site and output base
        stageout_location = task.crab_stageout_location()
        if not isinstance(stageout_location, (list, tuple)) or len(stageout_location) != 2:
            raise ValueError(
                "the return value of crab_stageout_location() is expected to be a 2-tuple, got "
                "'{}'".format(stageout_location),
            )
        c.storage_site, c.output_lfn_base = stageout_location

        # add the bootstrap file
        bootstrap_file = task.crab_bootstrap_file()
        if bootstrap_file:
            c.input_files["bootstrap_file"] = bootstrap_file

        # add the stageout file
        stageout_file = task.crab_stageout_file()
        if stageout_file:
            c.input_files["stageout_file"] = stageout_file

        # does the dashboard have a hook file?
        dashboard_file = self.dashboard.remote_hook_file()
        if dashboard_file:
            c.input_files["dashboard_file"] = dashboard_file

        # log file
        if task.transfer_logs:
            c.custom_log_file = "stdall.txt"

        # task hook
        c = task.crab_job_config(c, list(submit_jobs.keys()), list(submit_jobs.values()))

        # build the job file and get the sanitized config
        job_file, c = self.job_file_factory(**c.__dict__)

        # return job and log file entry
        # (the latter is None but will be synced from query data)
        return {"job": job_file, "config": c, "log": None}

    def _status_error_pairs(self, job_num, job_data):
        pairs = super(CrabWorkflowProxy, self)._status_error_pairs(job_num, job_data)

        # add site history
        pairs.insert_before("log", "site history", job_data["extra"].get("site_history", no_value))

        return pairs

    def destination_info(self):
        info = super(CrabWorkflowProxy, self).destination_info()

        info = self.task.crab_destination_info(info)

        return info


class CrabWorkflow(BaseRemoteWorkflow):

    workflow_proxy_cls = CrabWorkflowProxy

    crab_workflow_run_decorators = None
    crab_job_manager_defaults = None
    crab_job_file_factory_defaults = None

    crab_job_kwargs = []
    crab_job_kwargs_submit = None
    crab_job_kwargs_cancel = None
    crab_job_kwargs_cleanup = None
    crab_job_kwargs_query = None

    exclude_params_branch = set()
    exclude_params_crab_workflow = set()

    exclude_index = True

    @contextlib.contextmanager
    def crab_workflow_run_context(self):
        """
        Hook to provide a context manager in which the workflow run implementation is placed. This
        can be helpful in situations where resurces should be acquired before and released after
        running a workflow.
        """
        yield

    @abstractmethod
    def crab_stageout_location(self):
        """
        Hook to define both the "Site.storageSite" and "Data.outLFNDirBase" settings in a 2-tuple,
        i.e., the name of the storage site to use and the base directory for crab's own output
        staging. An example would be ``("T2_DE_DESY", "/store/user/...")``.

        In case this is not used, the choice of the output base has no affect, but is still required
        for crab's job submission to work.
        """
        return

    @abstractmethod
    def crab_output_directory(self):
        """
        Hook to define the location of submission output files, such as the json files containing
        job data. This method should return a :py:class:`FileSystemDirectoryTarget`.
        """
        return

    def crab_request_name(self, submit_jobs):
        """
        Returns a random name for a request, i.e., the project directory inside the crab job working
        area.
        """
        return "{}_{}".format(self.live_task_id, str(uuid.uuid4())[:8])

    def crab_work_area(self):
        """
        Returns the location of the crab working area, defaulting to the value of
        :py:meth:`crab_output_directory` in case it refers to a local directory. When *None*, the
        value of the "job.crab_work_area" configuration options is used.
        """
        # when job files are cleaned, try to use the output directory when local
        if self.workflow_proxy.job_file_factory and self.workflow_proxy.job_file_factory.cleanup:
            out_dir = self.crab_output_directory()
            # when local, return the directory
            if isinstance(out_dir, LocalDirectoryTarget):
                return out_dir
            # when not a target and no remote scheme, return the directory
            if (
                not isinstance(out_dir, FileSystemDirectoryTarget) and
                get_scheme(out_dir) in (None, "file")
            ):
                return remove_scheme(out_dir)

        # relative to the job file directory
        return ""

    def crab_job_file(self):
        """
        Hook to return the location of the job file that is executed on job nodes.
        """
        return JobInputFile(law_src_path("job", "law_job.sh"))

    def crab_bootstrap_file(self):
        """
        Hook to define the location of an optional, so-called bootstrap file that is sent alongside
        jobs and called prior to the actual job payload. It is meant to run a custom setup routine
        in order for the payload to run successfully (e.g. software setup, data retrieval).
        """
        return None

    def crab_stageout_file(self):
        """
        Hook to define the location of an optional, so-called stageout file that is sent alongside
        jobs and called after to the actual job payload. It is meant to run a custom output stageout
        routine if required so by your workflow or target storage element.
        """
        return None

    def crab_workflow_requires(self):
        """
        Hook to define requirements for the workflow itself and that need to be resolved before any
        submission can happen.
        """
        return DotDict()

    def crab_output_postfix(self):
        """
        Hook to define the postfix of outputs, for instance such that workflows with different
        parameters do not write their intermediate job status information into the same json file.
        """
        return ""

    def crab_output_uri(self):
        """
        Hook to return the URI of the remote crab output directory.
        """
        return self.crab_output_directory().uri()

    def crab_job_resources(self, job_num, branches):
        """
        Hook to define resources for a specific job with number *job_num*, processing *branches*.
        This method should return a dictionary.
        """
        return {}

    def crab_job_manager_cls(self):
        """
        Hook to define a custom job managet class to use.
        """
        return CrabJobManager

    def crab_create_job_manager(self, **kwargs):
        """
        Hook to configure how the underlying job manager is instantiated and configured.
        """
        kwargs = merge_dicts(self.crab_job_manager_defaults, kwargs)
        return self.crab_job_manager_cls()(**kwargs)

    def crab_job_file_factory_cls(self):
        """
        Hook to define a custom job file factory class to use.
        """
        return CrabJobFileFactory

    def crab_create_job_file_factory(self, **kwargs):
        """
        Hook to configure how the underlying job file factory is instantiated and configured.
        """
        # get the file factory cls
        factory_cls = self.crab_job_file_factory_cls()

        # job file fectory config priority: kwargs > class defaults
        kwargs = merge_dicts({}, self.crab_job_file_factory_defaults, kwargs)

        # default mkdtemp value which might require task-level info
        if kwargs.get("mkdtemp") is None:
            cfg = Config.instance()
            mkdtemp = cfg.get_expanded(
                "job",
                cfg.find_option("job", "crab_job_file_dir_mkdtemp", "job_file_dir_mkdtemp"),
            )
            if isinstance(mkdtemp, six.string_types) and mkdtemp.lower() not in {"true", "false"}:
                kwargs["mkdtemp"] = factory_cls._expand_template_path(
                    mkdtemp,
                    variables={"task_id": self.live_task_id, "task_family": self.task_family},
                )

        return factory_cls(**kwargs)

    def crab_job_config(self, config, submit_jobs):
        """
        Hook to inject custom settings into the job *config*, which is an instance of the
        :py:attr:`Config` class defined inside the job manager.
        """
        return config

    def crab_dump_intermediate_job_data(self):
        """
        Whether to dump intermediate job data to the job submission file while jobs are being
        submitted.
        """
        return True

    def crab_post_submit_delay(self):
        """
        Configurable delay in seconds to wait after submitting jobs and before starting the status
        polling.
        """
        return self.poll_interval * 60

    def crab_check_job_completeness(self):
        """
        Hook to define whether
        """
        return False

    def crab_check_job_completeness_delay(self):
        """
        Grace period before :py:meth:`crab_check_job_completeness` is called to ensure that output
        files are accessible. Especially useful on distributed file systems with possibly
        asynchronous behavior.
        """
        return 0.0

    def crab_poll_callback(self, poll_data):
        """
        Configurable callback that is called after each job status query and before potential
        resubmission. It receives the variable polling attributes *poll_data* (:py:class:`PollData`)
        that can be changed within this method.

        If *False* is returned, the polling loop is gracefully terminated. Returning any other value
        does not have any effect.
        """
        return

    def crab_cmdline_args(self):
        """
        Hook to add additional cli parameters to "law run" commands executed on job nodes.
        """
        return {}

    def crab_destination_info(self, info):
        """
        Hook to add additional information behind each job status query line by extending an *info*
        dictionary whose values will be shown separated by comma.
        """
        return info



# coding: utf-8

"""
Function returning the config defaults of the htcondor package.
"""


def config_defaults(default_config):
    return {
        "job": {
            "htcondor_job_grouping_submit": True,
            "htcondor_job_file_dir": None,
            "htcondor_job_file_dir_mkdtemp": None,
            "htcondor_job_file_dir_cleanup": False,
            "htcondor_chunk_size_submit": 25,
            "htcondor_chunk_size_cancel": 25,
            "htcondor_chunk_size_query": 25,
            "htcondor_merge_job_files": True,
        },
    }



# coding: utf-8

"""
"law sw" cli subprogram.
"""


import os
import sys
import shutil
from importlib import import_module

import six

from law.config import Config
from law.logger import get_logger


logger = get_logger(__name__)


# TODO: auto-detect
default_dep_names = ["six", "luigi", "law"]
if six.PY3:
    default_dep_names += ["tenacity", "dateutil"]

dep_names = os.getenv("LAW_SOFTWARE_DEPS", None)
if dep_names:
    dep_names = [name.strip() for name in dep_names.strip().split(",")]
else:
    dep_names = default_dep_names

_deps = None


if "_reloaded_deps" not in globals():
    _reloaded_deps = False


def setup_parser(sub_parsers):
    """
    Sets up the command line parser for the *software* subprogram and adds it to *sub_parsers*.
    """
    csv = lambda s: [_s.strip() for _s in str(s).strip().split(",")]

    parser = sub_parsers.add_parser(
        "software",
        prog="law software",
        description="Create or update the law software cache ({}). This is only required for some "
        "sandboxes that need to forward software into (e.g.) containers.".format(get_sw_dir()),
    )

    parser.add_argument(
        "--deps",
        "-d",
        type=csv,
        default=dep_names,
        help="comma-separated names of python packages to cache; "
        "default: {}".format(",".join(dep_names)),
    )
    parser.add_argument(
        "--remove",
        "-r",
        action="store_true",
        help="remove the software cache directory and exit",
    )
    parser.add_argument(
        "--location",
        "-l",
        action="store_true",
        help="print the location of the software cache directory and exit",
    )
    parser.add_argument(
        "--print-deps",
        "-p",
        action="store_true",
        help="print the list of dependencies and exit",
    )


def execute(args):
    """
    Executes the *software* subprogram with parsed commandline *args*.
    """
    sw_dir = get_sw_dir()

    # just print the cache location?
    if args.location:
        print(sw_dir)
        return

    # just print the list of dependencies?
    if args.print_deps:
        print(",".join(args.deps))
        return

    # just remove the current software cache?
    if args.remove:
        remove_software_cache(sw_dir)
        return

    # rebuild the software cache
    build_software_cache(sw_dir, dep_names=args.deps)


def get_software_deps(names=None):
    global _deps

    if _deps is not None:
        return _deps
    _deps = []

    if names is None:
        names = list(dep_names)

    for name in names:
        try:
            mod = import_module(name)
        except ImportError as e:
            print("could not import software dependency '{}': {}".format(name, e))
            continue

        _deps.append(mod)

    return _deps


def build_software_cache(sw_dir=None, dep_names=None):
    """
    Builds up the software cache directory at *sw_dir* by simply copying all required python
    modules identified by *dep_names*, defaulting to a predefined list of package names. *sw_dir*
    is evaluated with :py:func:`get_sw_dir`.
    """
    # ensure the cache is empty
    sw_dir = get_sw_dir(sw_dir)
    remove_software_cache(sw_dir)
    os.makedirs(sw_dir)

    # reload dependencies to find the proper module paths
    reload_dependencies(force=True)

    # get dependencies
    deps = get_software_deps(names=dep_names)

    for mod in deps:
        path = os.path.dirname(mod.__file__)
        name, ext = os.path.splitext(os.path.basename(mod.__file__))
        # single file or module?
        if name == "__init__":
            # copy the entire module
            name = os.path.basename(path)
            shutil.copytree(path, os.path.join(sw_dir, name))
        else:
            shutil.copy2(os.path.join(path, name + ".py"), sw_dir)
        logger.debug("cached '{}'".format(mod))


def remove_software_cache(sw_dir=None):
    """
    Removes the software cache directory at *sw_dir* which is evaluated with :py:func:`get_sw_dir`.
    """
    sw_dir = get_sw_dir(sw_dir)
    if os.path.exists(sw_dir):
        shutil.rmtree(sw_dir)


def reload_dependencies(force=False, dep_names=None):
    """
    Reloads all python modules that law depends on, idenfied by *dep_names* and defaulting to a
    predefined list of package names. Unless *force* is *True*, multiple calls to this function will
    not have any effect.
    """
    global _reloaded_deps

    # reload only once
    if _reloaded_deps and not force:
        return
    _reloaded_deps = True

    # get dependencies
    deps = get_software_deps(names=dep_names)

    # reload them
    for mod in deps:
        six.moves.reload_module(mod)
        logger.debug("reloaded '{}'".format(mod))


def use_software_cache(sw_dir=None, reload_deps=False):
    """
    Adjusts ``sys.path`` so that the cached software at *sw_dir* is used. *sw_dir* is evaluated with
    :py:func:`get_sw_dir`. When *reload_deps* is *True*, :py:func:`reload_dependencies` is invoked.
    """
    sw_dir = get_sw_dir(sw_dir)
    if not os.path.exists(sw_dir):
        return

    # ammend the search path to favor the software cache
    sys.path.insert(1, sw_dir)

    # reload dependencies in case they were already loaded
    if reload_deps:
        reload_dependencies()


def get_sw_dir(sw_dir=None):
    """
    Returns the software directory defined in the ``core.software_dir`` config. When *sw_dir* is not
    *None*, it is expanded and returned instead.
    """
    if sw_dir is None:
        cfg = Config.instance()
        sw_dir = cfg.get_expanded("core", "software_dir")

    sw_dir = os.path.expandvars(os.path.expanduser(sw_dir))

    return sw_dir



# coding: utf-8

"""
Docker sandbox implementation.
"""

__all__ = ["DockerSandbox"]


import os
import sys
import uuid
import socket
import subprocess

import luigi
import six

from law.config import Config
from law.sandbox.base import Sandbox
from law.target.local import LocalFileTarget
from law.cli.software import get_software_deps
from law.util import make_list, interruptable_popen, quote_cmd, flatten, makedirs


class DockerSandbox(Sandbox):

    sandbox_type = "docker"

    config_section_prefix = sandbox_type

    @property
    def image(self):
        return self.name

    @property
    def tag(self):
        return None if ":" not in self.image else self.image.split(":", 1)[1]

    @property
    def env_cache_key(self):
        return self.image

    def get_custom_config_section_postfix(self):
        return self.image

    def create_env(self):
        # strategy: create a tempfile, forward it to a container, let python dump its full env,
        # close the container and load the env file

        # helper to load the env
        def load_env(target):
            try:
                return tmp.load(formatter="pickle")
            except Exception as e:
                raise Exception(
                    "env deserialization of sandbox {} failed: {}".format(self, e),
                )

        # load the env when the cache file is configured and existing
        if self.env_cache_path:
            env_cache_target = LocalFileTarget(self.env_cache_path)
            if env_cache_target.exists():
                return load_env(env_cache_target)

        tmp = LocalFileTarget(is_tmp=".env")
        tmp.touch()

        env_file = os.path.join("/tmp", tmp.unique_basename)

        # get the docker run command
        docker_run_cmd = self._docker_run_cmd()

        # mount the env file
        docker_run_cmd.extend(["-v", "{}:{}".format(tmp.path, env_file)])

        # pre-setup commands
        pre_setup_cmds = self._build_pre_setup_cmds()

        # post-setup commands
        post_env = self._get_env()
        post_setup_cmds = self._build_post_setup_cmds(post_env)

        # build the python command that dumps the environment
        py_cmd = "import os,pickle;" \
            + "pickle.dump(dict(os.environ),open('{}','wb'),protocol=2)".format(env_file)

        # build the full command
        cmd = quote_cmd(docker_run_cmd + [self.image, "bash", "-l", "-c",
            " && ".join(flatten(
                pre_setup_cmds,
                post_setup_cmds,
                quote_cmd(["python", "-c", py_cmd]),
            )),
        ])

        # run it
        code, out, _ = interruptable_popen(cmd, shell=True, executable="/bin/bash",
            stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        if code != 0:
            raise Exception(
                "docker sandbox env loading failed with exit code {}:\n{}".format(
                    code, out),
            )

        # copy to the cache path when configured
        if self.env_cache_path:
            tmp.copy_to_local(env_cache_target)

        # load the env
        env = load_env(tmp)

        return env

    def _docker_run_cmd(self):
        """
        Part of the "docker run" command that is common to env requests and run.
        """
        cmd = ["docker", "run"]

        # rm flag
        cmd.extend(["--rm"])

        # use the pid namespace of the host so killing the outer process will stop the container
        cmd.extend(["--pid", "host"])

        # task-specific arguments
        if self.task:
            # user flag
            sandbox_user = self.task.sandbox_user()
            if sandbox_user:
                if not isinstance(sandbox_user, (tuple, list)) or len(sandbox_user) != 2:
                    raise Exception("sandbox_user() must return 2-tuple")
                cmd.extend(["-u", "{}:{}".format(*sandbox_user)])

            # add args configured on the task
            args_getter = getattr(self.task, "docker_args", None)
            if callable(args_getter):
                cmd.extend(make_list(args_getter()))

        return cmd

    def cmd(self, proxy_cmd):
        # docker run command arguments
        args = []

        # container name
        args.extend(["--name", "{}_{}".format(self.task.task_id, str(uuid.uuid4())[:8])])

        # container hostname
        args.extend(["-h", "{}".format(socket.gethostname())])

        # helper to build forwarded paths
        cfg = Config.instance()
        cfg_section = self.get_config_section()
        forward_dir = cfg.get_expanded(cfg_section, "forward_dir")
        python_dir = cfg.get_expanded(cfg_section, "python_dir")
        bin_dir = cfg.get_expanded(cfg_section, "bin_dir")
        stagein_dir_name = cfg.get_expanded(cfg_section, "stagein_dir_name")
        stageout_dir_name = cfg.get_expanded(cfg_section, "stageout_dir_name")

        def dst(*args):
            return os.path.join(forward_dir, *(str(arg) for arg in args))

        # helper for mounting a volume
        volume_srcs = []

        def mount(*vol):
            src = vol[0]

            # make sure, the same source directory is not mounted twice
            if src in volume_srcs:
                return
            volume_srcs.append(src)

            # ensure that source directories exist
            if not os.path.isfile(src):
                makedirs(src)

            # store the mount point
            args.extend(["-v", ":".join(vol)])

        # environment variables to set
        env = self._get_env()

        # add staging directories
        if self.stagein_info:
            env["LAW_SANDBOX_STAGEIN_DIR"] = dst(stagein_dir_name)
            mount(self.stagein_info.stage_dir.path, dst(stagein_dir_name))
        if self.stageout_info:
            env["LAW_SANDBOX_STAGEOUT_DIR"] = dst(stageout_dir_name)
            mount(self.stageout_info.stage_dir.path, dst(stageout_dir_name))

        # prevent python from writing byte code files
        env["PYTHONDONTWRITEBYTECODE"] = "1"

        # adjust path variables
        env["PATH"] = os.pathsep.join([dst("bin"), "$PATH"])
        env["PYTHONPATH"] = os.pathsep.join([dst(python_dir), "$PYTHONPATH"])

        # forward python directories of law and dependencies
        for mod in get_software_deps():
            path = os.path.dirname(mod.__file__)
            name, ext = os.path.splitext(os.path.basename(mod.__file__))
            if name == "__init__":
                vsrc = path
                vdst = dst(python_dir, os.path.basename(path))
            else:
                vsrc = os.path.join(path, name + ".py")
                vdst = dst(python_dir, name + ".py")
            mount(vsrc, vdst, "ro")

        # forward the law cli dir to bin as it contains a law executable
        env["PATH"] = os.pathsep.join([dst(python_dir, "law", "cli"), env["PATH"]])

        # forward the law config file
        if cfg.config_file:
            mount(cfg.config_file, dst("law.cfg"))
            env["LAW_CONFIG_FILE"] = dst("law.cfg")

        # forward the luigi config file
        for p in luigi.configuration.LuigiConfigParser._config_paths[::-1]:
            if os.path.exists(p):
                mount(p, dst("luigi.cfg"))
                env["LUIGI_CONFIG_PATH"] = dst("luigi.cfg")
                break

        # forward volumes defined in the config and by the task
        vols = self._get_volumes()
        for hdir, cdir in six.iteritems(vols):
            if not cdir:
                mount(hdir, hdir)
            else:
                cdir = self._expand_volume(cdir, bin_dir=dst(bin_dir), python_dir=dst(python_dir))
                mount(hdir, cdir)

        # handle local scheduling within the container
        if self.force_local_scheduler():
            proxy_cmd.add_arg("--local-scheduler", "True", overwrite=True)
        elif self.scheduler_on_host():
            # when the scheduler runs on the host system, we need to set the network interface to
            # the host system and set the correct host address as seen by the container
            args.extend(["--network", "host"])
            proxy_cmd.add_arg("--scheduler-host", self.get_host_ip(), overwrite=True)

        # get the docker run command, add arguments from above
        docker_run_cmd = self._docker_run_cmd() + args

        # pre-setup commands
        pre_setup_cmds = self._build_pre_setup_cmds()

        # post-setup commands with the full env
        post_setup_cmds = self._build_post_setup_cmds(env)

        # build the final command
        cmd = quote_cmd(docker_run_cmd + [self.image, "bash", "-l", "-c",
            " && ".join(flatten(
                pre_setup_cmds,
                post_setup_cmds,
                proxy_cmd.build(),
            )),
        ])

        return cmd

    def get_host_ip(self):
        # in host network mode, docker containers can normally be accessed via 127.0.0.1 on Linux
        # or via docker.for.mac.localhost on Mac (as of docker 17.06), however, in some cases it
        # might be required to use a different ip which can be set via an env variable
        default_ip = "docker.for.mac.localhost" if sys.platform == "darwin" else "127.0.0.1"
        return os.getenv("LAW_DOCKER_HOST_IP", default_ip)



# coding: utf-8
# flake8: noqa

"""
ROOT contrib functionality.
"""

__all__ = [
    "import_ROOT", "hadd_task",
    "ROOTFormatter", "ROOTNumpyFormatter", "ROOTPandasFormatter", "UprootFormatter", "GuardedTFile",
]


# provisioning imports
from law.contrib.root.util import import_ROOT, hadd_task
from law.contrib.root.formatter import (
    ROOTFormatter, ROOTNumpyFormatter, ROOTPandasFormatter, UprootFormatter, GuardedTFile,
)



# coding: utf-8

"""
Slack notifications.
"""

__all__ = ["notify_slack"]


import threading

import six

from law.config import Config
from law.util import escape_markdown
from law.logger import get_logger


logger = get_logger(__name__)


def notify_slack(title, content, attachment_color="#4bb543", short_threshold=40, token=None,
        channel=None, mention_user=None, **kwargs):
    """
    Sends a slack notification and returns *True* on success. The communication with the slack API
    might have some delays and is therefore handled by a thread. The format of the notification
    depends on *content*. If it is a string, a simple text notification is sent. Otherwise, it
    should be a dictionary whose fields are used to build a message attachment with two-column
    formatting.
    """
    # test import
    import_slack()

    cfg = Config.instance()

    # get default token and channel
    if not token:
        token = cfg.get_expanded("notifications", "slack_token")
    if not channel:
        channel = cfg.get_expanded("notifications", "slack_channel")

    if not token or not channel:
        logger.warning("cannot send Slack notification, token ({}) or channel ({}) empty".format(
            token, channel))
        return False

    # append the user to mention to the title
    # unless explicitly set to empty string
    mention_text = ""
    if mention_user is None:
        mention_user = cfg.get_expanded("notifications", "slack_mention_user")
    if mention_user:
        mention_text = " (@{})".format(escape_markdown(mention_user))

    # request data for the API call
    request = {
        "channel": channel,
        "as_user": True,
        "parse": "full",
    }

    # standard or attachment content?
    if isinstance(content, six.string_types):
        request["text"] = "{}{}\n\n{}".format(title, mention_text, content)
    else:
        # content is a dict, send its data as an attachment
        request["text"] = "{} {}".format(title, mention_text)
        request["attachments"] = at = {
            "color": attachment_color,
            "fields": [],
            "fallback": "{}{}\n\n".format(title, mention_text),
        }

        # fill the attachment fields and extend the fallback
        for key, value in content.items():
            at["fields"].append({
                "title": key,
                "value": value,
                "short": len(value) <= short_threshold,
            })
            at["fallback"] += "_{}_: {}\n".format(key, value)

    # extend by arbitrary kwargs
    request.update(kwargs)

    # threaded, non-blocking API communication
    thread = threading.Thread(target=_notify_slack, args=(token, request))
    thread.start()

    return True


def _notify_slack(token, request):
    import os
    import json
    import traceback

    slack, vslack = import_slack()

    try:
        # token might be a file
        token_file = os.path.expanduser(os.path.expandvars(token))
        if os.path.isfile(token_file):
            with open(token_file, "r") as f:
                token = f.read().strip()

        if "attachments" in request and not isinstance(request["attachments"], six.string_types):
            request["attachments"] = json.dumps([request["attachments"]])

        if vslack == 1:
            sc = slack.SlackClient(token)
            res = sc.api_call("chat.postMessage", **request)
        else:  # 2
            wc = slack.WebClient(token)
            res = wc.chat_postMessage(**request)

        if not res["ok"]:
            logger.warning("unsuccessful Slack API call: {}".format(res))
    except Exception as e:
        t = traceback.format_exc()
        logger.warning("could not send Slack notification: {}\n{}".format(e, t))


def import_slack():
    try:
        # slackclient 1.x
        import slackclient  # noqa: F401
        return slackclient, 1
    except ImportError:
        try:
            # slackclient 2.x
            import slack  # noqa: F401
            return slack, 2
        except ImportError as e:
            e.msg = "neither module 'slackclient' nor 'slack' found, " \
                "run 'pip install slackclient' to install them"
            raise



# coding: utf-8

"""
Pandas target formatters.
"""

__all__ = ["DataFrameFormatter"]


import os

from law.target.formatter import Formatter
from law.target.file import get_path
from law.logger import get_logger
from law.util import no_value


logger = get_logger(__name__)


class DataFrameFormatter(Formatter):

    name = "pandas"

    @classmethod
    def accepts(cls, path, mode):
        # still missing: excel, html, xml, latex, feather, orc, sql, stata, markdown, ...
        suffixes = (".csv", ".json", ".parquet", ".h5", ".hdf5", ".pickle", ".pkl")
        return get_path(path).endswith(suffixes)

    @classmethod
    def load(cls, path, *args, **kwargs):
        path = get_path(path)

        if path.endswith(".csv"):
            import pandas  # fmt: skip
            return pandas.read_csv(path, *args, **kwargs)

        if path.endswith(".json"):
            import pandas  # fmt: skip
            return pandas.read_json(path, *args, **kwargs)

        if path.endswith(".parquet"):
            import pandas  # fmt: skip
            return pandas.read_parquet(path, *args, **kwargs)

        if path.endswith((".h5", ".hdf5")):
            import pandas  # fmt: skip
            return pandas.read_hdf(path, *args, **kwargs)

        if path.endswith((".pickle", ".pkl")):
            import pandas  # fmt: skip
            return pandas.read_pickle(path, *args, **kwargs)

        suffix = os.path.splitext(path)[1]
        raise NotImplementedError(
            "suffix \"{}\" not implemented in DataFrameFormatter".format(suffix),
        )

    @classmethod
    def dump(cls, path, obj, *args, **kwargs):
        _path = get_path(path)
        perm = kwargs.pop("perm", no_value)

        if _path.endswith(".csv"):
            ret = obj.to_csv(_path, *args, **kwargs)

        elif _path.endswith(".json"):
            ret = obj.to_json(_path, *args, **kwargs)

        elif _path.endswith(".parquet"):
            ret = obj.to_parquet(_path, *args, **kwargs)

        elif _path.endswith((".h5", ".hdf5")):
            ret = obj.to_hdf(_path, *args, **kwargs)

        elif _path.endswith((".pickle", ".pkl")):
            ret = obj.to_pickle(_path, *args, **kwargs)

        else:
            suffix = os.path.splitext(_path)[1]
            raise NotImplementedError(
                "suffix \"{}\" not implemented in DataFrameFormatter".format(suffix),
            )

        if perm != no_value:
            cls.chmod(path, perm)

        return ret



# coding: utf-8

"""
WLCG remote file system and targets.
"""

__all__ = ["WLCGFileSystem", "WLCGTarget", "WLCGFileTarget", "WLCGDirectoryTarget"]


import six

import law
from law.target.remote import (
    RemoteFileSystem, RemoteTarget, RemoteFileTarget, RemoteDirectoryTarget,
)
from law.logger import get_logger


logger = get_logger(__name__)

law.contrib.load("gfal")


class WLCGFileSystem(RemoteFileSystem):

    file_interface_cls = law.gfal.GFALFileInterface

    def __init__(self, section=None, **kwargs):
        # read configs from section and combine them with kwargs to get the file system and
        # file interface configs
        section, fs_config, fi_config = self._init_configs(section, "default_wlcg_fs",
            "_wlcg_fs_defaults", kwargs)

        # store the config section
        self.config_section = section
        fs_config.setdefault("name", self.config_section)

        # base path is mandatory
        if not fi_config.get("base"):
            raise Exception("attribute 'base' must not be empty, set it either directly in the {} "
                "constructor, or add the option 'base' to your config section '{}'".format(
                    self.__class__.__name__, self.config_section))

        # enforce some configs
        fs_config["has_permissions"] = False

        # create the file interface
        file_interface = self.file_interface_cls(**fi_config)

        # initialize the file system itself
        super(WLCGFileSystem, self).__init__(file_interface, **fs_config)


# try to set the default fs instance
try:
    WLCGFileSystem.default_instance = WLCGFileSystem()
    logger.debug("created default WLCGFileSystem instance '{}'".format(
        WLCGFileSystem.default_instance))
except Exception as e:
    logger.debug("could not create default WLCGFileSystem instance: {}".format(e))


class WLCGTarget(RemoteTarget):
    """ __init__(path, fs=WLCGFileSystem.default_instance, **kwargs)
    """

    def __init__(self, path, fs=WLCGFileSystem.default_instance, **kwargs):
        if fs is None:
            fs = WLCGFileSystem.default_instance
        elif isinstance(fs, six.string_types):
            fs = WLCGFileSystem(fs)
        RemoteTarget.__init__(self, path, fs, **kwargs)


class WLCGFileTarget(WLCGTarget, RemoteFileTarget):

    pass


class WLCGDirectoryTarget(WLCGTarget, RemoteDirectoryTarget):

    pass


WLCGTarget.file_class = WLCGFileTarget
WLCGTarget.directory_class = WLCGDirectoryTarget



# coding: utf-8

"""
Interface for communicating with a remote file service.
"""

__all__ = ["RemoteFileInterface"]


import os
import sys
import time
import abc
import functools
import random as _random

import six

from law.config import Config
from law.target.file import remove_scheme
from law.util import make_list, is_lazy_iterable, brace_expand, parse_duration
from law.logger import get_logger


logger = get_logger(__name__)


class RetryException(Exception):

    def __init__(self, msg="", exc=None):
        self.exc_type, self.exc_value, self.exc_traceback = exc or sys.exc_info()
        super(RetryException, self).__init__(msg or str(self.exc_value))

    def reraise(self):
        return six.reraise(self.exc_type, self.exc_value, self.exc_traceback)


class RemoteFileInterface(six.with_metaclass(abc.ABCMeta, object)):

    @classmethod
    def parse_config(cls, section, config=None, overwrite=False):
        cfg = Config.instance()

        if config is None:
            config = {}

        # helper to add a config value if it exists, extracted with a config parser method
        def add(option, func):
            if option not in config or overwrite:
                config[option] = func(section, option)

        def get_expanded_list(section, option):
            # get config value, run brace expansion taking into account csv splitting
            value = cfg.get_expanded(section, option, None)
            return value and [v.strip() for v in brace_expand(value.strip(), split_csv=True)]

        def get_time(section, option):
            value = cfg.get_expanded(section, option)
            return parse_duration(value, input_unit="s", unit="s")

        # base path(s)
        add("base", get_expanded_list)

        # base path(s) per operation
        options = cfg.options(section, prefix="base_")
        add("bases", lambda *_: {
            option[5:]: get_expanded_list(section, option)
            for option in options
            if not cfg.is_missing_or_none(section, option)
        })

        # default number of retries
        add("retries", cfg.get_expanded_int)

        # default delay between retries
        add("retry_delay", get_time)

        # default setting for the random base selection
        add("random_base", cfg.get_expanded_bool)

        return config

    @classmethod
    def retry(cls, func=None, uri_cmd=None, uri_base_name=None):
        # cmd will be deprecated and fully renamed to base_name, but has priority for now
        if uri_cmd:
            logger.warning_once(
                "deprecate_RemoteFileInterface_retry_uri_cmd",
                "the argument 'uri_cmd' in {}.retry is deprected, ".format(cls.__name__) +
                "please use 'uri_base_name' instead",
            )
            uri_base_name = uri_cmd

        def decorator(func):
            @functools.wraps(func)
            def wrapper(self, *args, **kwargs):
                # function name for logs
                func_name = func.__name__

                # get retry configs with defaults from the interface itself
                retries = kwargs.pop("retries", self.retries)
                delay = kwargs.pop("retry_delay", self.retry_delay)
                random_base = kwargs.pop("random_base", self.random_base)

                attempt = 0
                try:
                    base = None
                    base_set = bool(kwargs.get("base"))
                    skip_indices = []
                    while True:
                        # when no base was set initially and a uri_base_name is given, get a random
                        # uri base under consideration of bases (given by their indices) to skip
                        if not base_set and uri_base_name:
                            base, idx = self.get_base(base_name=uri_base_name, random=random_base,
                                skip_indices=skip_indices, return_index=True)
                            kwargs["base"] = base
                            skip_indices.append(idx)

                        try:
                            return func(self, *args, **kwargs)
                        except RetryException as e:
                            attempt += 1

                            # raise to the outer try-except block when there are no attempts left
                            if attempt > retries:
                                e.reraise()

                            # log and sleep
                            logger.debug("{}.{}(args: {}, kwargs: {}) failed: {}, retry".format(
                                self.__class__.__name__, func_name, args, kwargs, e))
                            time.sleep(delay)
                except:
                    # at this point, no more retry attempts are available,
                    # so update the exception to reflect that, then reraise
                    e_type, e, traceback = sys.exc_info()
                    msg = str(e)
                    msg += "\nfunction: {}.{}".format(self.__class__.__name__, func_name)
                    msg += "\nattempts: {}".format(attempt)
                    msg += "\nargs    : {}".format(args)
                    msg += "\nkwargs  : {}".format(kwargs)
                    msg += "\nerror   : {}: '{}'".format(e_type.__name__, e)
                    six.reraise(e_type, e_type(msg, *e.args[1:]), traceback)

            return wrapper

        return decorator(func) if func else decorator

    def __init__(self, base=None, bases=None, retries=0, retry_delay=0, random_base=True, **kwargs):
        super(RemoteFileInterface, self).__init__()

        # convert base(s) to list for random selection
        base = make_list(base or [])
        bases = {k: make_list(b) for k, b in six.iteritems(bases)} if bases else {}

        # at least one base in expected
        if len(base) == 0:
            raise Exception("{} expected at least one base path, received none".format(
                self.__class__.__name__))

        # expand variables in base and bases
        expand = lambda p: os.path.expandvars(str(p))
        self.base = list(map(expand, base))
        self.bases = {k: list(map(expand, b)) for k, b in six.iteritems(bases)}

        # store other attributes
        self.retries = retries
        self.retry_delay = retry_delay
        self.random_base = random_base

    def sanitize_path(self, p):
        return str(p)

    def get_base(self, cmd=None, base_name=None, random=None, skip_indices=None, return_index=False,
            return_all=False):
        if random is None:
            random = self.random_base

        # cmd will be deprecated and fully renamed to base_name, but has priority for now
        if cmd:
            logger.warning_once(
                "deprecate_RemoteFileInterface_get_base_cmd",
                "the argument 'cmd' in {}.get_base is deprected, ".format(self.__class__.__name__) +
                "please use 'base_name' instead",
            )
            base_name = cmd

        # get potential bases for the given base_name
        bases = make_list(self.base)
        if base_name:
            for base_name in make_list(base_name):
                if base_name in self.bases:
                    bases = self.bases[base_name]
                    break

        if not bases:
            raise Exception("no bases available for command '{}'".format(base_name))

        # are there indices to skip?
        all_bases = bases
        if skip_indices:
            _bases = [b for i, b in enumerate(bases) if i not in skip_indices]
            if _bases:
                bases = _bases

        # return all?
        if return_all:
            return bases

        # select one
        if len(bases) == 1 or not random:
            # select the first base
            base = bases[0]
        else:
            # select a random base
            base = _random.choice(bases)

        return base if not return_index else (base, all_bases.index(base))

    def uri(self, path, base=None, return_all=False, scheme=True, **kwargs):
        # get a base path when not given
        if not base:
            kwargs["return_index"] = False
            base = self.get_base(return_all=return_all, **kwargs)

        # helper to join the path to some base b and remove the scheme if requested
        def uri(b):
            uri = os.path.join(b, self.sanitize_path(path).lstrip("/")).rstrip("/")
            return uri if scheme else remove_scheme(uri)

        if isinstance(base, (list, tuple)) or is_lazy_iterable(base):
            return [uri(b) for b in base]

        if return_all:
            return [uri(base)]

        return uri(base)

    @abc.abstractmethod
    def exists(self, path, base=None, stat=False, **kwargs):
        """
        Returns *True* when the *path* exists and *False* otherwise. When *stat* is *True*, returns
        the stat object or *None*.
        """
        return

    @abc.abstractmethod
    def stat(self, path, base=None, **kwargs):
        """
        Returns a stat object or raises an exception when *path* does not exist.
        """
        return

    @abc.abstractmethod
    def isdir(self, path, stat=None, base=None, **kwargs):
        """
        Returns *True* when *path* refers to an existing directory, optionally using a precomputed
        stat object instead, and *False* otherwise.
        """
        return

    @abc.abstractmethod
    def isfile(self, path, stat=None, base=None, **kwargs):
        """
        Returns *True* when *path* refers to a existing file, optionally using a precomputed stat
        object instead, and *False* otherwise.
        """
        return

    @abc.abstractmethod
    def chmod(self, path, perm, base=None, silent=False, **kwargs):
        """
        Changes the permission of a *path* to *perm*. Raises an exception when *path* does not exist
        or returns *False* when *silent* is *True*, and returns *True* on success.
        """
        return

    @abc.abstractmethod
    def unlink(self, path, base=None, silent=True, **kwargs):
        """
        Removes a file at *path*. Raises an exception when *path* does not exist or returns *False*
        when *silent* is *True*, and returns *True* on success.
        """
        return

    @abc.abstractmethod
    def rmdir(self, path, base=None, silent=True, **kwargs):
        """
        Removes a directory at *path*. Raises an exception when *path* does not exist or returns
        *False* when *silent* is *True*, and returns *True* on success.
        """
        return

    @abc.abstractmethod
    def remove(self, path, base=None, silent=True, **kwargs):
        """
        Removes any file or directory at *path*. Directories are removed recursively. Raises an
        exception when *path* does not exist or returns *False* when *silent* is *True*, and returns
        *True* on success.
        """
        return

    @abc.abstractmethod
    def mkdir(self, path, perm, base=None, silent=True, **kwargs):
        """
        Creates a directory at *path* with permissions *perm*. Raises an exception when *path*
        already exists or returns *False* when *silent* is *True*, and returns *True* on success.
        """
        return

    @abc.abstractmethod
    def mkdir_rec(self, path, perm, base=None, **kwargs):
        """
        Recursively creates a directory and intermediate missing directories at *path* with
        permissions *perm*. Raises an exception when *path* already exists or returns *False* when
        *silent* is *True*, and returns *True* on success.
        """
        return

    @abc.abstractmethod
    def listdir(self, path, base=None, **kwargs):
        """
        Returns a list of elements in and relative to *path*.
        """
        return

    @abc.abstractmethod
    def filecopy(self, src, dst, base=None, **kwargs):
        """
        Copies a file from *src* to *dst*. Returns the full, schemed *src* and *dst* URIs used for
        copying in a 2-tuple.
        """
        return



# coding: utf-8

"""
"law completion" cli subprogram.
"""


from law.util import law_src_path


def setup_parser(sub_parsers):
    """
    Sets up the command line parser for the *completion* subprogram and adds it to *sub_parsers*.
    """
    sub_parsers.add_parser(
        "completion",
        prog="law completion",
        description="Print the location of the law completion file to source it in your shell via "
        "'source $(law completion)'.",
    )


def execute(args):
    """
    Executes the *completion* subprogram with parsed commandline *args*.
    """
    print(law_src_path("cli", "completion.sh"))



# coding: utf-8

"""
LSF job manager. See https://www.ibm.com/support/knowledgecenter/en/SSETD4_9.1.3.
"""

__all__ = ["LSFJobManager", "LSFJobFileFactory"]


import os
import stat
import time
import re
import subprocess

import six

from law.config import Config
from law.job.base import BaseJobManager, BaseJobFileFactory, JobInputFile, DeprecatedInputFiles
from law.target.file import get_path
from law.util import interruptable_popen, make_list, make_unique, quote_cmd
from law.logger import get_logger

from law.contrib.lsf.util import get_lsf_version


logger = get_logger(__name__)

_cfg = Config.instance()


class LSFJobManager(BaseJobManager):

    # chunking settings
    chunk_size_submit = 0
    chunk_size_cancel = _cfg.get_expanded_int("job", "lsf_chunk_size_cancel")
    chunk_size_query = _cfg.get_expanded_int("job", "lsf_chunk_size_query")

    submission_job_id_cre = re.compile(r"^Job <(\d+)> is submitted.+$")

    def __init__(self, queue=None, emails=False, threads=1):
        super(LSFJobManager, self).__init__()

        self.queue = queue
        self.emails = emails
        self.threads = threads

        # determine the LSF version once
        self.lsf_version = get_lsf_version()

        # flags for versions with some important changes
        self.lsf_v912 = self.lsf_version and self.lsf_version >= (9, 1, 2)

    def cleanup(self, *args, **kwargs):
        raise NotImplementedError("LSFJobManager.cleanup is not implemented")

    def cleanup_batch(self, *args, **kwargs):
        raise NotImplementedError("LSFJobManager.cleanup_batch is not implemented")

    def submit(self, job_file, queue=None, emails=None, retries=0, retry_delay=3, silent=False):
        # default arguments
        if queue is None:
            queue = self.queue
        if emails is None:
            emails = self.emails

        # get the job file location as the submission command is run it the same directory
        job_file_dir, job_file_name = os.path.split(os.path.abspath(str(job_file)))

        # build the command
        cmd = "LSB_JOB_REPORT_MAIL={} bsub".format("Y" if emails else "N")
        if queue:
            cmd += " -q {}".format(queue)
        cmd += " < {}".format(job_file_name)

        # define the actual submission in a loop to simplify retries
        while True:
            # run the command
            logger.debug("submit lsf job with command '{}'".format(cmd))
            code, out, err = interruptable_popen(cmd, shell=True, executable="/bin/bash",
                stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=job_file_dir)

            # get the job id
            if code == 0:
                m = self.submission_job_id_cre.match(out.strip())
                if m:
                    job_id = m.group(1)
                else:
                    code = 1
                    err = "cannot parse job id from output:\n{}".format(out)

            # retry or done?
            if code == 0:
                return job_id

            logger.debug("submission of lsf job '{}' failed with code {}:\n{}".format(
                job_file, code, err))

            if retries > 0:
                retries -= 1
                time.sleep(retry_delay)
                continue

            if silent:
                return None

            raise Exception("submission of lsf job '{}' failed: \n{}".format(job_file, err))

    def cancel(self, job_id, queue=None, silent=False):
        # default arguments
        if queue is None:
            queue = self.queue

        chunking = isinstance(job_id, (list, tuple))
        job_ids = make_list(job_id)

        # build the command
        cmd = ["bkill"]
        if queue:
            cmd += ["-q", queue]
        cmd += job_ids
        cmd = quote_cmd(cmd)

        # run it
        logger.debug("cancel lsf job(s) with command '{}'".format(cmd))
        code, out, err = interruptable_popen(cmd, shell=True, executable="/bin/bash",
            stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        # check success
        if code != 0 and not silent:
            raise Exception("cancellation of lsf job(s) '{}' failed with code {}:\n{}".format(
                job_id, code, err))

        return {job_id: None for job_id in job_ids} if chunking else None

    def query(self, job_id, queue=None, silent=False):
        # default arguments
        if queue is None:
            queue = self.queue

        chunking = isinstance(job_id, (list, tuple))
        job_ids = make_list(job_id)

        # build the command
        cmd = ["bjobs"]
        if self.lsf_v912:
            cmd.append("-noheader")
        if queue:
            cmd += ["-q", queue]
        cmd += job_ids
        cmd = quote_cmd(cmd)

        # run it
        logger.debug("query lsf job(s) with command '{}'".format(cmd))
        code, out, err = interruptable_popen(cmd, shell=True, executable="/bin/bash",
            stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        # handle errors
        if code != 0:
            if silent:
                return None
            else:
                raise Exception("status query of lsf job(s) '{}' failed with code {}:\n{}".format(
                    job_id, code, err))

        # parse the output and extract the status per job
        query_data = self.parse_query_output(out)

        # compare to the requested job ids and perform some checks
        for _job_id in job_ids:
            if _job_id not in query_data:
                if not chunking:
                    if silent:
                        return None
                    else:
                        raise Exception("lsf job(s) '{}' not found in query response".format(
                            job_id))
                else:
                    query_data[_job_id] = self.job_status_dict(job_id=_job_id, status=self.FAILED,
                        error="job not found in query response")

        return query_data if chunking else query_data[job_id]

    @classmethod
    def parse_query_output(cls, out):
        """
        Example output to parse:
        141914132 user_name DONE queue_name exec_host b63cee711a job_name Feb 8 14:54
        """
        query_data = {}

        for line in out.strip().split("\n"):
            parts = line.split()
            if len(parts) < 6:
                continue

            job_id = parts[0]
            status_flag = parts[2]

            # map the status
            status = cls.map_status(status_flag)

            # save the result
            query_data[job_id] = cls.job_status_dict(job_id=job_id, status=status)

        return query_data

    @classmethod
    def map_status(cls, status_flag):
        # https://www.ibm.com/support/knowledgecenter/en/SSETD4_9.1.2/lsf_command_ref/bjobs.1.html
        if status_flag in ("PEND", "PROV", "PSUSP", "USUSP", "SSUSP", "WAIT"):
            return cls.PENDING
        elif status_flag in ("RUN",):
            return cls.RUNNING
        elif status_flag in ("DONE",):
            return cls.FINISHED
        elif status_flag in ("EXIT", "UNKWN", "ZOMBI"):
            return cls.FAILED
        else:
            return cls.FAILED


class LSFJobFileFactory(BaseJobFileFactory):

    config_attrs = BaseJobFileFactory.config_attrs + [
        "file_name", "command", "executable", "arguments", "queue", "cwd", "input_files",
        "output_files", "postfix_output_files", "manual_stagein", "manual_stageout", "job_name",
        "stdout", "stderr", "shell", "emails", "custom_content", "absolute_paths",
    ]

    def __init__(self, file_name="lsf_job.job", command=None, executable=None, arguments=None,
            queue=None, cwd=None, input_files=None, output_files=None, postfix_output_files=True,
            manual_stagein=False, manual_stageout=False, job_name=None, stdout="stdout.txt",
            stderr="stderr.txt", shell="bash", emails=False, custom_content=None,
            absolute_paths=False, **kwargs):
        # get some default kwargs from the config
        cfg = Config.instance()
        if kwargs.get("dir") is None:
            kwargs["dir"] = cfg.get_expanded("job", cfg.find_option("job",
                "lsf_job_file_dir", "job_file_dir"))
        if kwargs.get("mkdtemp") is None:
            kwargs["mkdtemp"] = cfg.get_expanded_bool("job", cfg.find_option("job",
                "lsf_job_file_dir_mkdtemp", "job_file_dir_mkdtemp"))
        if kwargs.get("cleanup") is None:
            kwargs["cleanup"] = cfg.get_expanded_bool("job", cfg.find_option("job",
                "lsf_job_file_dir_cleanup", "job_file_dir_cleanup"))

        super(LSFJobFileFactory, self).__init__(**kwargs)

        self.file_name = file_name
        self.command = command
        self.executable = executable
        self.arguments = arguments
        self.queue = queue
        self.cwd = cwd
        self.input_files = DeprecatedInputFiles(input_files or {})
        self.output_files = output_files or []
        self.postfix_output_files = postfix_output_files
        self.manual_stagein = manual_stagein
        self.manual_stageout = manual_stageout
        self.job_name = job_name
        self.stdout = stdout
        self.stderr = stderr
        self.shell = shell
        self.emails = emails
        self.custom_content = custom_content
        self.absolute_paths = absolute_paths

    def create(self, postfix=None, **kwargs):
        # merge kwargs and instance attributes
        c = self.get_config(**kwargs)

        # some sanity checks
        if not c.file_name:
            raise ValueError("file_name must not be empty")
        if not c.command and not c.executable:
            raise ValueError("either command or executable must not be empty")
        if not c.shell:
            raise ValueError("shell must not be empty")

        # ensure that the custom log file is an output file
        if c.custom_log_file and c.custom_log_file not in c.output_files:
            c.output_files.append(c.custom_log_file)

        # postfix certain output files
        c.output_files = list(map(str, c.output_files))
        if c.postfix_output_files:
            skip_postfix_cre = re.compile(r"^(/dev/).*$")
            skip_postfix = lambda s: bool(skip_postfix_cre.match(str(s)))
            c.output_files = [
                path if skip_postfix(path) else self.postfix_output_file(path, postfix)
                for path in c.output_files
            ]
            for attr in ["stdout", "stderr", "custom_log_file"]:
                if c[attr] and not skip_postfix(c[attr]):
                    c[attr] = self.postfix_output_file(c[attr], postfix)

        # ensure that all input files are JobInputFile objects
        c.input_files = {
            key: JobInputFile(f)
            for key, f in c.input_files.items()
        }

        # ensure that the executable is an input file, remember the key to access it
        if c.executable:
            executable_keys = [
                k
                for k, v in c.input_files.items()
                if get_path(v) == get_path(c.executable)
            ]
            if executable_keys:
                executable_key = executable_keys[0]
            else:
                executable_key = "executable_file"
                c.input_files[executable_key] = JobInputFile(c.executable)

        # prepare input files
        def prepare_input(f):
            # when not copied, just return the absolute, original path
            abs_path = os.path.abspath(f.path)
            if not f.copy or f.forward:
                return abs_path
            # copy the file
            abs_path = self.provide_input(
                src=abs_path,
                postfix=postfix if f.postfix and not f.share else None,
                dir=c.dir,
                skip_existing=f.share,
            )
            return abs_path

        # absolute input paths
        for key, f in c.input_files.items():
            f.path_sub_abs = prepare_input(f)

        # input paths relative to the submission or initial dir
        # forwarded files are skipped as they are not treated as normal inputs
        for key, f in c.input_files.items():
            if f.forward:
                continue
            f.path_sub_rel = (
                os.path.basename(f.path_sub_abs)
                if f.copy and not c.absolute_paths else
                f.path_sub_abs
            )

        # input paths as seen by the job, before and after potential rendering
        for key, f in c.input_files.items():
            f.path_job_pre_render = (
                f.path_sub_abs
                if f.forward else
                os.path.basename(f.path_sub_abs)
            )
            f.path_job_post_render = (
                f.path_sub_abs
                if f.forward and not f.render_job else
                os.path.basename(f.path_sub_abs)
            )

        # update files in render variables with version after potential rendering
        c.render_variables.update({
            key: f.path_job_post_render
            for key, f in c.input_files.items()
        })

        # add space separated input files before potential rendering to render variables
        c.render_variables["input_files"] = " ".join(
            f.path_job_pre_render
            for f in c.input_files.values()
        )

        # add space separated list of input files for rendering
        c.render_variables["input_files_render"] = " ".join(
            f.path_job_pre_render
            for f in c.input_files.values()
            if f.render_job
        )

        # add the custom log file to render variables
        if c.custom_log_file:
            c.render_variables["log_file"] = c.custom_log_file

        # add the file postfix to render variables
        if postfix and "file_postfix" not in c.render_variables:
            c.render_variables["file_postfix"] = postfix

        # linearize render variables
        render_variables = self.linearize_render_variables(c.render_variables)

        # prepare the job description file
        job_file = self.postfix_input_file(os.path.join(c.dir, str(c.file_name)), postfix)

        # render copied, non-forwarded input files
        for key, f in c.input_files.items():
            if not f.copy or f.forward or not f.render_local:
                continue
            self.render_file(
                f.path_sub_abs,
                f.path_sub_abs,
                render_variables,
                postfix=postfix if f.postfix else None,
            )

        # prepare the executable when given
        if c.executable:
            c.executable = get_path(c.input_files[executable_key].path_job_post_render)
            # make the file executable for the user and group
            path = os.path.join(c.dir, os.path.basename(c.executable))
            if os.path.exists(path):
                os.chmod(path, os.stat(path).st_mode | stat.S_IXUSR | stat.S_IXGRP)

        # job file content
        content = []
        content.append("#!/usr/bin/env {}".format(c.shell))

        if c.job_name:
            content.append(("-J", c.job_name))
        if c.queue:
            content.append(("-q", c.queue))
        if c.cwd:
            content.append(("-cwd", c.cwd))
        if c.stdout:
            content.append(("-o", c.stdout))
        if c.stderr:
            content.append(("-e", c.stderr))
        if c.emails:
            content.append(("-N",))
        if c.custom_content:
            content += c.custom_content

        if not c.manual_stagein:
            paths = [f.path_sub_rel for f in c.input_files.values() if f.path_sub_rel]
            for path in make_unique(paths):
                content.append(("-f", "\"{} > {}\"".format(path, os.path.basename(path))))

        if not c.manual_stageout:
            for path in make_unique(c.output_files):
                content.append(("-f", "\"{} < {}\"".format(path, os.path.basename(path))))

        if c.manual_stagein:
            tmpl = "cp " + ("{}" if c.absolute_paths else "$LS_EXECCWD/{}") + " $PWD/{}"
            paths = [f.path_sub_rel for f in c.input_files.values() if f.path_sub_rel]
            for path in make_unique(paths):
                content.append(tmpl.format(path, os.path.basename(path)))

        if c.command:
            content.append(c.command)
        else:
            content.append("./" + c.executable)
        if c.arguments:
            args = quote_cmd(c.arguments) if isinstance(c.arguments, (list, tuple)) else c.arguments
            content[-1] += " {}".format(args)

        if c.manual_stageout:
            tmpl = "cp $PWD/{} $LS_EXECCWD/{}"
            for path in c.output_files:
                content.append(tmpl.format(path, path))

        # write the job file
        with open(job_file, "w") as f:
            for line in content:
                if not isinstance(line, six.string_types):
                    line = self.create_line(*make_list(line))
                f.write(line + "\n")

        logger.debug("created lsf job file at '{}'".format(job_file))

        return job_file, c

    @classmethod
    def create_line(cls, key, value=None):
        if value is None:
            return "#BSUB {}".format(key)
        else:
            return "#BSUB {} {}".format(key, value)



# coding: utf-8

"""
Implementation of a file interface using GFAL.
"""

__all__ = ["GFALFileInterface"]


import os
import sys
import contextlib
import stat as _stat

import six

from law.config import Config
from law.target.file import has_scheme, get_scheme
from law.target.remote.interface import RemoteFileInterface, RetryException
from law.logger import get_logger


logger = get_logger(__name__)


# try to import gfal2
try:
    import gfal2

    HAS_GFAL2 = True

except (ImportError, TypeError):
    HAS_GFAL2 = False

    class GFAL2Dummy(object):

        def __getattr__(self, attr):
            raise Exception("trying to access 'gfal2.{}', but gfal2 is not installed".format(attr))

    gfal2 = GFAL2Dummy()


class GFALFileInterface(RemoteFileInterface):

    @classmethod
    def parse_config(cls, section, config=None, overwrite=False):
        config = super(GFALFileInterface, cls).parse_config(section, config=config,
            overwrite=overwrite)

        cfg = Config.instance()

        # helper to add a config value if it exists, extracted with a config parser method
        def add(option, func, postfix="gfal_", _config=config):
            if option not in config or overwrite:
                _config[option] = func(section, postfix + option)

        # use atomic contexts per operation
        add("atomic_contexts", cfg.get_expanded_bool)

        # transfer config
        config.setdefault("transfer_config", {})
        transfer_specs = [
            ("timeout", cfg.get_expanded_int),
            ("checksum_check", cfg.get_expanded_bool),
            ("nbstreams", cfg.get_expanded_int),
            ("overwrite", cfg.get_expanded_bool),
            ("create_parent", cfg.get_expanded_bool),
            ("strict_copy", cfg.get_expanded_bool),
        ]
        for name, func in transfer_specs:
            add(name, func, "gfal_transfer_", config["transfer_config"])

        return config

    def __init__(self, atomic_contexts=False, gfal_options=None, transfer_config=None, **kwargs):
        super(GFALFileInterface, self).__init__(**kwargs)

        # cache for gfal context objects and transfer parameters per pid for thread safety
        self._contexts = {}
        self._transfer_parameters = {}

        # store gfal options and transfer configs
        self.gfal_options = gfal_options or {}
        self.transfer_config = transfer_config or {}

        # other configs
        self.atomic_contexts = atomic_contexts

    def sanitize_path(self, p):
        # in python 2, the gfal2-bindings do not support unicode but expect strings
        return str(p) if isinstance(p, six.string_types) else p

    @contextlib.contextmanager
    def context(self):
        # context objects are stored per pid, so create one if it does not exist yet
        pid = os.getpid()

        if pid not in self._contexts:
            self._contexts[pid] = ctx = gfal2.creat_context()
            for _type, args_list in six.iteritems(self.gfal_options):
                for args in args_list:
                    getattr(ctx, "set_opt_" + _type)(*args)

        # yield and optionally close it which frees potentially open connections
        try:
            yield self._contexts[pid]
        finally:
            if self.atomic_contexts and pid in self._contexts:
                del self._contexts[pid]

    @contextlib.contextmanager
    def transfer_parameters(self, ctx):
        pid = os.getpid()

        if pid not in self._transfer_parameters:
            self._transfer_parameters[pid] = ctx.transfer_parameters()
            for key, value in six.iteritems(self.transfer_config):
                setattr(self._transfer_parameters[pid], key, value)

        try:
            yield self._transfer_parameters[pid]
        finally:
            if self.atomic_contexts and pid in self._transfer_parameters:
                del self._transfer_parameters[pid]

    def exists(self, path, stat=False, base=None, **kwargs):
        uri = self.uri(path, base_name="stat" if stat else ("exists", "stat"), base=base)
        with self.context() as ctx:
            try:
                logger.debug("invoking gfal2 exists({})".format(uri))
                rstat = ctx.stat(uri)
                return rstat if stat else True
            except gfal2.GError:
                return None if stat else False

    @RemoteFileInterface.retry(uri_base_name="stat")
    def stat(self, path, base=None, **kwargs):
        uri = self.uri(path, base_name="stat", base=base)
        with self.context() as ctx:
            try:
                logger.debug("invoking gfal2 stat({})".format(uri))
                return ctx.stat(uri)

            except gfal2.GError:
                raise RetryException()

    def isdir(self, path, stat=None, base=None, **kwargs):
        if not stat:
            stat = self.exists(path, stat=True, base=base)

        if not stat:
            return False

        # some file protocols do not return standard st_mode values in stat requests,
        # e.g. srm returns file type bits 0o50000 for directories instead of 0o40000,
        # these differences are rather distinct and can be taken into account here,
        # see http://man7.org/linux/man-pages/man7/inode.7.html for info on st_mode values
        return _stat.S_ISDIR(stat.st_mode) or _stat.S_IFMT(stat.st_mode) == 0o50000

    def isfile(self, path, stat=None, base=None, **kwargs):
        if not stat:
            stat = self.exists(path, stat=True, base=base)

        if not stat:
            return False

        return not self.isdir(path, stat=stat, base=base)

    @RemoteFileInterface.retry(uri_base_name="chmod")
    def chmod(self, path, perm, base=None, silent=False, **kwargs):
        if perm is None:
            return True

        uri = self.uri(path, base_name="chmod", base=base)
        with self.context() as ctx:
            try:
                logger.debug("invoking gfal2 chmod({}, {})".format(uri, perm))
                ctx.chmod(uri, perm)
                return True

            except gfal2.GError:
                e = GFALError_chmod(uri)
                # check if the operation should be retried, can fail silently, or raised immediately
                if e.reason == e.UNKNOWN:
                    raise e
                if e.reason in (e.NOT_FOUND, e.NOT_SUPPORTED) and silent:
                    return False
                e.reraise()

    @RemoteFileInterface.retry(uri_base_name="unlink")
    def unlink(self, path, base=None, silent=True, **kwargs):
        uri = self.uri(path, base_name="unlink", base=base)
        with self.context() as ctx:
            try:
                logger.debug("invoking gfal2 unlink({})".format(uri))
                ctx.unlink(uri)
                return True

            except gfal2.GError:
                e = GFALError_unlink(uri)
                # check if the operation should be retried, can fail silently, or raised immediately
                if e.reason == e.UNKNOWN:
                    raise e
                if e.reason == e.NOT_FOUND and silent:
                    return False
                e.reraise()

    @RemoteFileInterface.retry(uri_base_name="rmdir")
    def rmdir(self, path, base=None, silent=True, **kwargs):
        uri = self.uri(path, base_name="rmdir", base=base)
        with self.context() as ctx:
            try:
                logger.debug("invoking gfal2 rmdir({})".format(uri))
                ctx.rmdir(uri)
                return True

            except gfal2.GError:
                e = GFALError_rmdir(uri)
                # check if the operation should be retried, can fail silently, or raised immediately
                if e.reason == e.UNKNOWN:
                    raise e
                if e.reason == e.NOT_FOUND and silent:
                    return False
                e.reraise()

    @RemoteFileInterface.retry(uri_base_name="unlink")
    def remove(self, path, base=None, silent=True, **kwargs):
        """
        Recursive removal is potentially expensive in terms of remote file operations, so this
        method is designed to reduce them as much as possible.
        """
        # most common use case is file removal, so try this first and in case there is an error
        # interpret its message to get more info on the object without further operations
        uri = self.uri(path, base_name="unlink", base=base)
        with self.context() as ctx:
            try:
                logger.debug("invoking gfal2 unlink({})".format(uri))
                ctx.unlink(uri)
                return True

            except gfal2.GError:
                e = GFALError_unlink(uri)
                # handle all cases, except when uri is a directory
                if e.reason != e.IS_DIRECTORY:
                    if e.reason == e.UNKNOWN:
                        raise e
                    if e.reason == e.NOT_FOUND and silent:
                        return False
                    e.reraise()

        # at this point, we are dealing with a directory so try to delete it right away
        uri = self.uri(path, base_name="rmdir", base=base)
        with self.context() as ctx:
            try:
                logger.debug("invoking gfal2 rmdir({})".format(uri))
                ctx.rmdir(uri)
                return True

            except gfal2.GError:
                e = GFALError_rmdir(uri)
                # handle all cases, except when the directory is not empty
                if e.reason != e.NOT_EMPTY:
                    if e.reason == e.UNKNOWN:
                        raise e
                    if e.reason == e.NOT_FOUND and silent:
                        return False
                    e.reraise()

        # the directory is not empty, so there is no other way than deleting contents recursively
        # first, and then removing the directory itself
        path = str(path)
        for elem in self.listdir(path, base=base, retries=0):
            self.remove(os.path.join(path, elem), base=base, silent=silent, retries=0)

        return self.rmdir(path, base=base, silent=silent, retries=0)

    @RemoteFileInterface.retry(uri_base_name="mkdir")
    def mkdir(self, path, perm, base=None, silent=True, **kwargs):
        uri = self.uri(path, base_name="mkdir", base=base)
        with self.context() as ctx:
            try:
                logger.debug("invoking gfal2 mkdir({}, {})".format(uri, perm))
                ctx.mkdir(uri, perm)
                return True

            except gfal2.GError:
                e = GFALError_mkdir(uri)
                # check if the operation should be retried, can fail silently, or raised immediately
                if e.reason == e.UNKNOWN:
                    raise e
                if e.reason == e.EXISTS and silent:
                    # fail silently only when uri is really a dictionary
                    if self.isdir(path, base=base):
                        return False
                e.reraise()

    @RemoteFileInterface.retry(uri_base_name=["mkdir_rec", "mkdir"])
    def mkdir_rec(self, path, perm, base=None, silent=True, **kwargs):
        uri = self.uri(path, base_name="mkdir", base=base)
        with self.context() as ctx:
            try:
                logger.debug("invoking gfal2 mkdir_rec({}, {})".format(uri, perm))
                ctx.mkdir_rec(uri, perm)
                return True

            except gfal2.GError:
                e = GFALError_mkdir(uri)
                # check if the operation should be retried, can fail silently, or raised immediately
                if e.reason == e.UNKNOWN:
                    raise e
                if e.reason == e.EXISTS and silent:
                    # fail silently only when uri is really a dictionary
                    if self.isdir(path, base=base):
                        return False
                e.reraise()

    @RemoteFileInterface.retry(uri_base_name="listdir")
    def listdir(self, path, base=None, **kwargs):
        uri = self.uri(path, base_name="listdir", base=base)
        with self.context() as ctx:
            try:
                logger.debug("invoking gfal2 listdir({})".format(uri))
                return ctx.listdir(uri)

            except gfal2.GError:
                e = GFALError_listdir(uri)
                # some protocols throw an error upon listdir on empty directories
                if e.EMPTY:
                    return []
                e.reraise()

    @RemoteFileInterface.retry(uri_base_name="filecopy")
    def filecopy(self, src, dst, base=None, **kwargs):
        if has_scheme(src):
            src_uri = self.sanitize_path(src)
        else:
            src_uri = self.uri(src, base_name="filecopy", base=base)

        if has_scheme(dst):
            dst_uri = self.sanitize_path(dst)
        else:
            dst_uri = self.uri(dst, base_name="filecopy", base=base)

        with self.context() as ctx, self.transfer_parameters(ctx) as params:
            try:
                logger.debug("invoking gfal2 filecopy({}, {})".format(src_uri, dst_uri))
                ctx.filecopy(params, src_uri, dst_uri)

            except gfal2.GError:
                e = GFALError_filecopy(src_uri, dst_uri)
                # check if the operation should be retried or raised immediately
                if e.reason == e.UNKNOWN:
                    raise e
                e.reraise()

        return src_uri, dst_uri


class GFALOperationError(RetryException):

    UNKNOWN = "unknown reason"

    def __init__(self, uri, exc=None):
        # store uri and scheme
        self.uri = str(uri)
        self.scheme = get_scheme(uri)

        # get the original error objects and find the error reason
        exc = exc or sys.exc_info()
        self.reason = self._get_reason(str(exc[1]), self.uri, self.scheme)

        # add the error reason to the message
        msg = "{} ({}: {})".format(exc[1], self.__class__.__name__, self.reason)

        super(GFALOperationError, self).__init__(msg=msg, exc=exc)

    @classmethod
    def _get_reason(cls, msg, uri, scheme):
        raise NotImplementedError()


class GFALError_chmod(GFALOperationError):

    NOT_FOUND = "no such file or directory"
    NOT_SUPPORTED = "chmod operation not supported"

    @classmethod
    def _get_reason(cls, msg, uri, scheme):
        lmsg = msg.lower()
        if scheme == "root":
            if "no such file or directory" in lmsg:
                return cls.NOT_FOUND

        elif scheme == "gsiftp":
            if "no such file or directory" in lmsg:
                return cls.NOT_FOUND

        elif scheme == "srm":
            if "no such file or directory" in lmsg:
                return cls.NOT_FOUND
            if "operation not supported" in lmsg:
                return cls.NOT_SUPPORTED

        elif scheme in ("dav", "davs"):
            if "protocol not supported" in lmsg:
                return cls.NOT_SUPPORTED

        elif scheme == "dropbox":
            if "protocol not supported" in lmsg:
                return cls.NOT_SUPPORTED

        else:
            logger.warning("scheme '{}' not known to {}, cannot parse '{}'".format(
                scheme, cls.__name__, msg))

        return cls.UNKNOWN


class GFALError_unlink(GFALOperationError):

    NOT_FOUND = "target not found"
    IS_DIRECTORY = "target is a directory"

    @classmethod
    def _get_reason(cls, msg, uri, scheme):
        lmsg = msg.lower()
        if scheme == "root":
            if "no such file or directory" in lmsg:
                return cls.NOT_FOUND
            if "is a directory" in lmsg:
                return cls.IS_DIRECTORY

        elif scheme == "gsiftp":
            if "no such file or directory" in lmsg:
                return cls.NOT_FOUND
            if "not a file" in lmsg:
                return cls.IS_DIRECTORY

        elif scheme == "srm":
            if "no such file" in lmsg:
                return cls.NOT_FOUND
            if "not a file" in lmsg:
                return cls.IS_DIRECTORY

        elif scheme in ("dav", "davs"):
            if "file not found" in lmsg:
                return cls.NOT_FOUND
            if "is a directory" in lmsg:
                return cls.IS_DIRECTORY

        elif scheme == "dropbox":
            if "not_found" in lmsg:
                return cls.NOT_FOUND

        else:
            logger.warning("scheme '{}' not known to {}, cannot parse '{}'".format(
                scheme, cls.__name__, msg))

        return cls.UNKNOWN


class GFALError_rmdir(GFALOperationError):

    NOT_FOUND = "target not found"
    IS_FILE = "target is a file"
    NOT_EMPTY = "directory is not empty"

    @classmethod
    def _get_reason(cls, msg, uri, scheme):
        lmsg = msg.lower()
        if scheme == "root":
            if "no such file or directory" in lmsg:
                return cls.NOT_FOUND
            if "not a directory" in lmsg:
                return cls.IS_FILE
            if "no such device" in lmsg:
                # cryptic message for non-empty directory
                return cls.NOT_EMPTY

        elif scheme == "gsiftp":
            if "no such file or directory" in lmsg:
                return cls.NOT_FOUND
            if "not a directory" in lmsg:
                return cls.IS_FILE
            if "directory is not empty" in lmsg:
                return cls.NOT_EMPTY

        elif scheme == "srm":
            if "no such file or directory" in lmsg:
                return cls.NOT_FOUND
            if "this file is not a directory" in lmsg:
                return cls.IS_FILE
            if "directory not empty" in lmsg:
                return cls.NOT_EMPTY

        elif scheme in ("dav", "davs"):
            if "file not found" in lmsg:
                return cls.NOT_FOUND

        elif scheme == "dropbox":
            if "not_found" in lmsg:
                return cls.NOT_FOUND

        else:
            logger.warning("scheme '{}' not known to {}, cannot parse '{}'".format(
                scheme, cls.__name__, msg))

        return cls.UNKNOWN


class GFALError_mkdir(GFALOperationError):

    EXISTS = "target already exists"

    @classmethod
    def _get_reason(cls, msg, uri, scheme):
        lmsg = msg.lower()
        if scheme == "root":
            if "file exists" in lmsg:
                return cls.EXISTS

        elif scheme == "gsiftp":
            if "directory already exists" in lmsg:
                return cls.EXISTS

        elif scheme == "srm":
            if "directory already exist" in lmsg:
                return cls.EXISTS

        elif scheme == "dropbox":
            if "the directory already exists" in lmsg:
                return cls.EXISTS

        else:
            logger.warning("scheme '{}' not known to {}, cannot parse '{}'".format(
                scheme, cls.__name__, msg))

        return cls.UNKNOWN


class GFALError_listdir(GFALOperationError):

    EMPTY = "directory is empty"

    @classmethod
    def _get_reason(cls, msg, uri, scheme):
        lmsg = msg.lower()
        if scheme == "root":
            # xrootd throws an expcetion when a directory is empty
            if lmsg.strip().endswith("invalid response (unknown error 303)"):
                return cls.EMPTY

        return cls.UNKNOWN


class GFALError_filecopy(GFALOperationError):

    SRC_NOT_FOUND = "source not found"
    DST_EXISTS = "target already exists"

    def __init__(self, src_uri, dst_uri, exc=None):
        # store uri and scheme
        self.src_uri = str(src_uri)
        self.dst_uri = str(dst_uri)
        self.src_scheme = get_scheme(src_uri)
        self.dst_scheme = get_scheme(dst_uri)

        # get the original error objects and find the error reason
        exc = exc or sys.exc_info()
        self.reason = self._get_reason(str(exc[1]), self.src_uri, self.dst_uri, self.src_scheme,
            self.dst_scheme)

        # add the error reason to the message
        msg = "{} ({}: {})".format(exc[1], self.__class__.__name__, self.reason)

        # bypass the GFALOperationError init
        RetryException.__init__(self, msg=msg, exc=exc)

    @classmethod
    def _get_reason(cls, msg, src_uri, dst_uri, src_scheme, dst_scheme):
        # in gfal, error messages on missing source files or existing target files depend on both
        # source and destination protocols, so all cases need to be handled separately
        lmsg = msg.lower()
        if (src_scheme, dst_scheme) == ("file", "file"):
            if "could not open source" in lmsg:
                return cls.SRC_NOT_FOUND
            if "the file exists" in lmsg:
                return cls.DST_EXISTS

        elif (src_scheme, dst_scheme) == ("file", "root"):
            if "no such file or directory (source)" in lmsg:
                return cls.SRC_NOT_FOUND
            if "file exists (destination)" in lmsg:
                return cls.DST_EXISTS

        elif (src_scheme, dst_scheme) == ("file", "gsiftp"):
            if "local system call no such file or directory" in lmsg:
                return cls.SRC_NOT_FOUND
            if "file exists" in lmsg:
                return cls.DST_EXISTS

        elif (src_scheme, dst_scheme) == ("file", "srm"):
            if "local system call no such file or directory" in lmsg:
                return cls.SRC_NOT_FOUND
            if "file exists" in lmsg:
                return cls.DST_EXISTS

        elif (src_scheme, dst_scheme) in (("file", "dav"), ("file", "davs")):
            if "local system call no such file or directory" in lmsg:
                return cls.SRC_NOT_FOUND

        elif (src_scheme, dst_scheme) == ("file", "dropbox"):
            if "could not open source" in lmsg:
                return cls.SRC_NOT_FOUND
            if "the file exists" in lmsg:
                return cls.DST_EXISTS

        elif (src_scheme, dst_scheme) == ("gsiftp", "file"):
            if "no such file or directory on url" in lmsg:
                return cls.SRC_NOT_FOUND
            if "the file exists" in lmsg:
                return cls.DST_EXISTS

        elif (src_scheme, dst_scheme) == ("gsiftp", "gsiftp"):
            if "file not found" in lmsg:
                return cls.SRC_NOT_FOUND
            if "destination already exist" in lmsg:
                return cls.DST_EXISTS

        elif (src_scheme, dst_scheme) in (("gsiftp", "dav"), ("gsiftp", "davs")):
            if "is a directory" in lmsg:
                return cls.DST_EXISTS

        elif (src_scheme, dst_scheme) == ("srm", "file"):
            if "no such file" in lmsg:
                return cls.SRC_NOT_FOUND
            if "the file exists" in lmsg:
                return cls.DST_EXISTS

        elif (src_scheme, dst_scheme) == ("srm", "root"):
            if "no such file" in lmsg:
                return cls.SRC_NOT_FOUND
            if "the file exists" in lmsg:
                return cls.DST_EXISTS

        elif (src_scheme, dst_scheme) == ("srm", "srm"):
            if "no such file" in lmsg:
                return cls.SRC_NOT_FOUND
            if "file exists" in lmsg:
                return cls.DST_EXISTS

        elif (src_scheme, dst_scheme) in (("srm", "dav"), ("srm", "davs")):
            if "no such file" in lmsg:
                return cls.SRC_NOT_FOUND
            if "file exists" in lmsg:
                return cls.DST_EXISTS

        elif (src_scheme, dst_scheme) == ("root", "file"):
            if "no such file or directory" in lmsg:
                return cls.SRC_NOT_FOUND
            if "file exists (destination)" in lmsg:
                return cls.DST_EXISTS

        elif (src_scheme, dst_scheme) == ("root", "srm"):
            if "no such file or directory" in lmsg:
                return cls.SRC_NOT_FOUND
            if "file exists" in lmsg:
                return cls.DST_EXISTS

        elif (src_scheme, dst_scheme) == ("root", "root"):
            if "destination does not support delegation." in lmsg:
                return cls.SRC_NOT_FOUND
            if "file exists" in lmsg:
                return cls.DST_EXISTS

        elif (src_scheme, dst_scheme) in (("root", "dav"), ("root", "davs")):
            if "failed to open file (block device required)" in lmsg:
                return cls.DST_EXISTS

        elif (src_scheme, dst_scheme) in (("dav", "file"), ("davs", "file")):
            if "file not found" in lmsg:
                return cls.SRC_NOT_FOUND

        elif (src_scheme, dst_scheme) in (("dav", "gsiftp"), ("davs", "gsiftp")):
            if "file not found" in lmsg:
                return cls.SRC_NOT_FOUND
            if "not a file" in lmsg:
                return cls.DST_EXISTS

        elif (src_scheme, dst_scheme) in (("dav", "root"), ("davs", "root")):
            # it appears that there is a bug in gfal when copying via davix to xrootd in that
            # the full dst path is repeated, e.g. "root://url.tld:1090/pnfs/.../root://url..."
            # which causes weird behavior, and as long as this issue persists, there should be no
            # error parsing in law
            pass

        elif (src_scheme, dst_scheme) in (("dav", "srm"), ("davs", "srm")):
            # same issue as for davix -> xrootd, wait until this is resolved
            pass

        elif (src_scheme, dst_scheme) == ("dropbox", "file"):
            if "could not open source" in lmsg:
                return cls.SRC_NOT_FOUND
            if "the file exists" in lmsg:
                return cls.DST_EXISTS

        elif (src_scheme, dst_scheme) == ("dropbox", "dropbox"):
            if "could not open source" in lmsg:
                return cls.SRC_NOT_FOUND
            if "the file exists" in lmsg:
                return cls.DST_EXISTS

        else:
            logger.warning("combination of source scheme '{}' and target scheme '{}' not known to "
                "{}, cannot parse '{}'".format(src_scheme, dst_scheme, cls.__name__, msg))

        return cls.UNKNOWN



# coding: utf-8

"""
Function returning the config defaults of the arc package.
"""


def config_defaults(default_config):
    return {
        "job": {
            "arc_job_file_dir": None,
            "arc_job_file_dir_mkdtemp": None,
            "arc_job_file_dir_cleanup": None,
            "arc_chunk_size_submit": 25,
            "arc_chunk_size_cancel": 25,
            "arc_chunk_size_cleanup": 25,
            "arc_chunk_size_query": 20,
        },
    }



# coding: utf-8
# flake8: noqa

"""
HTCondor contrib functionality.
"""

__all__ = [
    "get_htcondor_version",
    "HTCondorJobManager", "HTCondorJobFileFactory",
    "HTCondorWorkflow",
]


# provisioning imports
from law.contrib.htcondor.util import get_htcondor_version
from law.contrib.htcondor.job import HTCondorJobManager, HTCondorJobFileFactory
from law.contrib.htcondor.workflow import HTCondorWorkflow



# coding: utf-8

"""
Custom base target definition.
"""

__all__ = ["Target"]


from abc import abstractmethod

from law.config import Config
import law.target.luigi_shims as shims
from law.util import colored, create_hash
from law.logger import get_logger


logger = get_logger(__name__)


class Target(shims.Target):

    def __init__(self, **kwargs):
        self.optional = kwargs.pop("optional", False)
        self.external = kwargs.pop("external", False)

        super(Target, self).__init__(**kwargs)

    def __repr__(self):
        color = Config.instance().get_expanded_bool("target", "colored_repr")
        return self.repr(color=color)

    def __str__(self):
        color = Config.instance().get_expanded_bool("target", "colored_str")
        return self.repr(color=color)

    def __hash__(self):
        return self.hash

    @property
    def hash(self):
        return create_hash(self.uri(), to_int=True)

    def repr(self, color=None):
        if color is None:
            color = Config.instance().get_expanded_bool("target", "colored_repr")

        class_name = self._repr_class_name(self.__class__.__name__, color=color)

        parts = [self._repr_pair(*pair, color=color) for pair in self._repr_pairs()]
        parts += [self._repr_flag(flag, color=color) for flag in self._repr_flags()]

        return "{}({})".format(class_name, ", ".join(parts))

    def colored_repr(self):
        # deprecation warning until v0.1
        logger.warning_once("the use of {0}.colored_repr() is deprecated, please use "
            "{0}.repr(color=True) instead".format(self.__class__.__name__))

        return self.repr(color=True)

    def _repr_pairs(self):
        return []

    def _repr_flags(self):
        flags = []
        if self.optional:
            flags.append("optional")
        if self.external:
            flags.append("external")
        return flags

    def _repr_class_name(self, name, color=False):
        return colored(name, "cyan") if color else name

    def _repr_pair(self, key, value, color=False):
        return "{}={}".format(colored(key, color="blue", style="bright") if color else key, value)

    def _repr_flag(self, name, color=False):
        return colored(name, color="magenta") if color else name

    def _copy_kwargs(self):
        return {"optional": self.optional, "external": self.external}

    def status_text(self, max_depth=0, flags=None, color=False, exists=None):
        if exists is None:
            exists = self.exists()

        if exists:
            text = "existent"
            _color = "green"
        else:
            text = "absent"
            _color = "grey" if self.optional else "red"

        return colored(text, _color, style="bright") if color else text

    def complete(self, **kwargs):
        """
        Returns almost the same state information as :py:meth:`exists` (called internally), but
        potentially also includes settings such as :py:attr:`optional`. All *kwargs* are forwarded
        to :py:meth:`exists`.

        This method is mostly useful in conjunction with task implementations whereas the vanilla
        :py:meth:`exists` method should be used when relying on the actual existence status.
        """
        return self.optional or self.exists(**kwargs)

    @abstractmethod
    def exists(self):
        return

    @abstractmethod
    def remove(self, silent=True):
        return

    @abstractmethod
    def uri(self, return_all=False):
        return
